/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "interp.h"
#include "QHObjAssocArr.h"
#include "timer.h"
#if defined(sgi)
#include <strings.h>
#endif

#define STATIC_CELL_NUM_OBJECTS 256
#define RESIZE_CELL_SIZE 10000

#ifndef MAX_CELL_SIZE
#define MAX_CELL_SIZE 200000
#endif

static CELL* extended_cell_list = NULL;
static CELL* free_cell_list = NULL;
static CELL  static_cell_objects[STATIC_CELL_NUM_OBJECTS];
Bool OK_DO_garbage_collecting = False;
int count_of_garbage_collection = 0;
static int number_of_expanded_cell = 0;
static void mark_object(QHObj* obj);
static int ncollect;

#define NUMBER_OF_DATA_BLOCK 2048
struct data_block* data_block_ptr = NULL;

#define IS_STATIC_CELL_OBJECT(p) ((static_cell_objects <= (p)) && \
		  ((p) < static_cell_objects + STATIC_CELL_NUM_OBJECTS))

/* for debug
static void print_markd_object(void)
{
    int i;
    CELL* p;

    for(i = 0; i < STATIC_CELL_NUM_OBJECTS; i++)
    {
	if(static_cell_objects[i].header.mark)
	{
	    printf("*");
	}
	else
	    printf(".");
    }
    putchar('\n');

    for(p = extended_cell_list; p != NULL; p = p->header.next)
    {
	if(p->header.mark)
	{
	    printf("*");
	}
	else
	    printf(".");
    }
    putchar('\n');
}
*/

static Bool is_cell_type(int type)
{
    return type == QH_Array ||
	   type == QH_AssocArray ||
           type == QH_String ||
           type == QH_Struct ||
           type == QH_List;
}

static void free_cell_memory(CELL* p);
static void free_cell_memory_for_assoc_array(const char* name, QHObj* value)
{
    if(is_cell_type(value->type))
	free_cell_memory(value->data.cell);
}

static void free_cell_memory(CELL* p)
{
    if(p == NULL)
	return;
    if(p->header.type == QH_Array) /* array */
    {
	p->header.type = None;
	if(!p->header.shared_data)
	{
	    if(p->a.type == QH_Any)
	    {
		int i;
		QHObj* x = p->a.elem;
		for(i = 0; i < p->a.nelem; i++)
		    if(is_cell_type(x[i].type))
		       free_cell_memory(x[i].data.cell);
	    }
	    if(p->a.dbp == NULL)
	    {
		xfree(p->a.size);
		xfree(p->a.elem);
	    }
	}
    }
    else if(p->header.type == QH_AssocArray)
    {
	p->header.type = None;
	mapQHObjAssocArr(p->assoc.tab, free_cell_memory_for_assoc_array);
	deleteQHObjAssocArr(p->assoc.tab);
	xfree(p->assoc.tab);
    }
    else if(p->header.type == QH_String)
    {
	p->header.type = None;
	if(!p->header.shared_data && p->s.dbp == NULL)
	    xfree(p->s.str);
    }
    else if(p->header.type == QH_Struct)
    {
	p->header.type = None;
	if(p->st.dbp == NULL)
	    xfree(p->st.member);
    }
    else if(p->header.type == QH_ExtData)
    {
	p->header.type = None;
	if(p->ext.destructor != NULL)
	    p->ext.destructor(&p->ext);
    }
}

void reuse_data_block(struct data_block* block)
{
    block->next = data_block_ptr;
    data_block_ptr = block;
}

static void init_data_block(void)
{
    struct data_block* block;
    int i;
    block = (struct data_block *)
	xmalloc(NUMBER_OF_DATA_BLOCK * sizeof(struct data_block));

    for(i = 0; i < NUMBER_OF_DATA_BLOCK - 1; i++)
	block[i].next = &block[i + 1];
    block[i].next = NULL;
    data_block_ptr = block;
}

void init_cell_object(void)
{
    int i;

    bzero((char *)static_cell_objects, sizeof(static_cell_objects));
    extended_cell_list = NULL;
    free_cell_list = static_cell_objects;
    for(i = 0; i < STATIC_CELL_NUM_OBJECTS - 1; i++)
    {
	static_cell_objects[i].header.next = &static_cell_objects[i + 1];
	static_cell_objects[i].header.enext = NULL;
    }
    static_cell_objects[i].header.next = NULL;
    static_cell_objects[i].header.enext = NULL;
    init_data_block();
}

void free_cell_object(void)
{
    int i;
    CELL* p;
    CELL* tmp;

    p = extended_cell_list;
    while(p)
    {
	tmp = p;
	p = p->header.enext;
	free_cell_memory(tmp);
	xfree(tmp);
    }

    p = free_cell_list;
    while(p)
    {
	tmp = p;
	p = p->header.next;
	if(!IS_STATIC_CELL_OBJECT(tmp))
	{
	    free_cell_memory(tmp);
	    xfree(tmp);
	}
    }

    for(i = 0; i < STATIC_CELL_NUM_OBJECTS; i++)
	free_cell_memory(static_cell_objects + i);
}

/* a->type must be QH_Any */
static void mark_hetero_array_object(Array* a)
{
    int i;
    QHObj* x = a->elem;

    for(i = 0; i < a->nelem; i++)
	mark_object(x + i);
}

static void mark_assoc_array_object(const char* name, QHObj* value);
static void mark_object(QHObj* obj)
{
    if(obj->type == QH_Array)
    {
	if(obj->data.a->header.mark == False)
	{
	    obj->data.a->header.mark = True;
	    if(obj->data.a->type == QH_Any)
		mark_hetero_array_object(obj->data.a);
	}
    }
    else if(obj->type == QH_AssocArray)
    {
	if(obj->data.assoc->header.mark == False)
	{
	    obj->data.assoc->header.mark = True;
	    mapQHObjAssocArr(obj->data.assoc->tab, mark_assoc_array_object);
	}
    }
    else if(obj->type == QH_String)
    {
	obj->data.s->header.mark = True;
    }
    else if(obj->type == QH_Struct)
    {
	if(obj->data.st->header.mark == False)
	{
	    int i, n;

	    obj->data.st->header.mark = True;
	    n = struct_tag_table[obj->data.st->entry].nmember;
	    for(i = 0; i < n; i++)
		mark_object(obj->data.st->member + i);
	}
    }
    else if(obj->type == QH_List)
    {
	if(obj->data.list->header.mark == False)
	{
	    QHObj cur;

	    for(cur = *obj; cur.type == QH_List; cur = cur.data.list->cdr)
	    {
		cur.data.list->header.mark = True;
		mark_object(&cur.data.list->car);
	    }
	}
    }
    else if(obj->type == QH_ExtData)
    {
	if(obj->data.ext->header.mark == False)
	{
	    obj->data.ext->header.mark = True;
	    if(obj->data.ext->marker != NULL)
	    {
		obj->data.ext->marker(obj->data.ext);
	    }
	}
    }
}

static void mark_assoc_array_object(const char* name, QHObj* value)
{
    mark_object(value);
}

static void mark_all_object(void)
{
    int i;

    for(i = 0; i < qh_top; i++)
	mark_object(memstack + i);
}

static Bool delete_cell_object(CELL* p)
{
    switch(p->header.type)
    {
      case QH_Array:
	p->header.type = None;
	if(p->a.dbp != NULL)
	    reuse_data_block(p->a.dbp);
	else if(!p->header.shared_data)
	{
	    xfree(p->a.size);
	    xfree(p->a.elem);
	}
	return True;
      case QH_AssocArray:
	p->header.type = None;
	deleteQHObjAssocArr(p->assoc.tab);
	return True;
      case QH_String:
	p->header.type = None;
	if(p->s.dbp != NULL)
	    reuse_data_block(p->s.dbp);
	else if(!p->header.shared_data)
	    xfree(p->s.str);
	return True;
      case QH_Struct:
	p->header.type = None;
	if(p->st.dbp == NULL)
	    xfree(p->st.member);
	return True;
      case QH_List:
	p->header.type = None;
	return True;
      case QH_ExtData:
	p->header.type = None;
	if(p->ext.destructor != NULL)
	    p->ext.destructor(&p->ext);
	return True;
    }
    /* other object */
    /* None-objects don't delete */
    return False;
}


void garbage_collect(void)
{
    int i;
    CELL* p;
    CELL* cell;
    CELL* prev_cell;
    double rtime = 0, utime = 0;
    int nstatic_cell = 0;
    int nexpand_cell = 0;
    Bool flag;

    if(OK_DO_garbage_collecting == False)
    {
	fprintf(stderr, "Garbage collection isn't allowed.\n");
	fprintf(stderr, "Maybe the memory or the static objects are exhausted.\n");
	qhexit(1);
    }

    if(print_garbage_collection)
    {
	puts("#########################");
	printf("# beginning of garbace collection,  No.%d\n",
	       count_of_garbage_collection);
	printf("# pc = %d, opcode = `%s'\n", qh_pc, opstr(qhprog[qh_pc].op));
	puts("#########################");
	utime = get_laptime();
	rtime = get_current_calender_time();
    }

    /* free unused memory */
    free_memory_in_compiling_at_once();

    /* mark all object */
    mark_all_object();

/*    print_markd_object(); */

    /* collect unmarked object in static-cell */

    free_cell_list = NULL;
    for(i = 0; i < STATIC_CELL_NUM_OBJECTS; i++)
    {
	if(static_cell_objects[i].header.mark == True)
	{
	    static_cell_objects[i].header.mark = False;
	}
	else
	{
	    p = static_cell_objects + i;
	    if(delete_cell_object(p) == True)
	    {
		/* append p to free_cell_list; */
		p->header.next = free_cell_list;
		free_cell_list = p;
		nstatic_cell++;
	    }
	}
    }

    /* collect unmarked object in extended-cell */
    prev_cell = cell = NULL;
    flag = True;
    for(p = extended_cell_list; p != NULL;)
    {
	if(flag)
	    prev_cell = cell;
	cell = p;
	p = p->header.enext;

	if(cell->header.mark == True)
	{
	    cell->header.mark = False;
	    flag = True;
	}
	else
	{
	    if(delete_cell_object(cell) == True)
	    {
		flag = False;
		if(cell == extended_cell_list)
		{
		    extended_cell_list = extended_cell_list->header.enext;
		}
		else
		{
		    prev_cell->header.enext = p;
		}

		/* append cell to free_cell_list; */
		cell->header.next = free_cell_list;
		free_cell_list = cell;
		nexpand_cell++;
	    }
	    else
		flag = False;
	}
    }

    ncollect = nstatic_cell + nexpand_cell;

    if(print_garbage_collection)
    {
	
	puts("######################");
	printf("# finished [%d/%d (%.1f%%) cells available.]\n",
	       ncollect,
	       STATIC_CELL_NUM_OBJECTS + number_of_expanded_cell,
	       (double)ncollect * 100.0
	       		/ (STATIC_CELL_NUM_OBJECTS + number_of_expanded_cell)
	       );
	printf("# static cell: %d/%d, expand cell: %d/%d\n",
	       nstatic_cell, STATIC_CELL_NUM_OBJECTS,
	       nexpand_cell, number_of_expanded_cell);

	printf("# user: %.2f    real: %.2f\n",
	       get_laptime() - utime,
	       get_current_calender_time() - rtime);
	puts("######################");
	puts("");
    }
    count_of_garbage_collection++;
}

CELL* alloc_cell(void)
{
    CELL* p;

    if(free_cell_list == NULL)
    {
	garbage_collect();

	if(free_cell_list == NULL)
	{
	    /* try expand CELL space, and do garbage-collection again */
	    if(print_garbage_collection)
		fprintf(stderr,	"# Could not collect any CELL. "
			"Then expand CELL space if possible.\n");

	    OK_DO_garbage_collecting = False;
	    expand_CELL_space(RESIZE_CELL_SIZE);
	    OK_DO_garbage_collecting = True;
	    if(free_cell_list == NULL)
		execute_error("All CELL are exhausted");

	    /* free_cell_list is not NULL, becase new CELL is appended.
	     * So the follow recusive call must be successful.
	     */
	    p = alloc_cell();
	}
	else if(ncollect < 256)
	    expand_CELL_space(1024);
    }

    p = free_cell_list;
    free_cell_list = free_cell_list->header.next;

    p->header.type = None;
    p->header.mark = False;
    p->header.next = NULL;
    p->header.enext = NULL;
    p->header.shared_data = False;

    if(!IS_STATIC_CELL_OBJECT(p)) /* append `p' to extended_cell_list */
    {
	p->header.enext = extended_cell_list;
	extended_cell_list = p;
    }

    return p;
}

int expand_CELL_space(int n)
{
    int i;
    CELL* p;

#ifdef MAX_CELL_SIZE
    if(number_of_expanded_cell + n > MAX_CELL_SIZE - STATIC_CELL_NUM_OBJECTS)
	n = MAX_CELL_SIZE - STATIC_CELL_NUM_OBJECTS - number_of_expanded_cell;
#endif

    if(n <= 0)
	return 0;
    
    for(i = 0; i < n; i++)
    {
	p = (CELL *)xmalloc(sizeof(CELL));

	p->header.type = None;
	p->header.mark = False;
	p->header.enext = NULL;
	p->header.shared_data = False;
	p->header.next = free_cell_list;
	free_cell_list = p;
    }

    if(print_garbage_collection)
	printf("# expand %d cells.\n", n);

    number_of_expanded_cell += n;

    return n;
}
