/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "interp.h"
#include "functab.h"

typedef struct _funcall_table_t
{
    char* fname;		/* filename */
    int   narg;			/* number of arguments */
    int   f;			/* entry for memfunctab */
} funcall_table_t;

static funcall_table_t* funcall_table = NULL;
static int funcall_hash_size = 0;

static unsigned int hash(const char* str)
{
     unsigned long h = 0;

    while(*str)
    {
	h = ((h << 3) + (h >> 28)) + *str;
	str++;
    }

    return h % funcall_hash_size;
}   

static void install_funcall_table(int f)
{
    int addr;

    if(memfunctab[f].module_name == NULL)
	strcpy(temp_buff, memfunctab[f].name);
    else
	sprintf(temp_buff, "%s::%s", memfunctab[f].module_name, memfunctab[f].name);

    addr = hash(temp_buff);

    while(funcall_table[addr].fname != NULL)
	addr = (addr + 1) % funcall_hash_size;

    funcall_table[addr].fname = strdup(temp_buff);
    funcall_table[addr].narg  = memfunctab[f].narg;
    funcall_table[addr].f     = f;
}

static void undefined_function(const char* fname, int narg)
{
    sprintf(temp_buff, "funcall: Undefined function `%s/%d'", fname, narg);
    execute_error(temp_buff);
}    

int search_funcall_table(const char* fname, int narg)
{
    int addr;
    int start_addr;
    int orig_narg = narg;

    if(funcall_hash_size == 0)
	undefined_function(fname, narg);

    start_addr = addr = hash(fname);

  retry_once:
    while(1)
    {
	if(funcall_table[addr].fname == NULL)
	{
	    if(start_addr >= 0)
	    {
		addr = start_addr;
		start_addr = -1;
		narg = -1;
		goto retry_once;
	    }
/* Debug for funcall
	    for(i = 0; i < funcall_hash_size; i++)
		if(funcall_table[i].fname != NULL)
		    printf("[%d] %s/%d\n", i, funcall_table[i].fname, funcall_table[i].narg);
*/
	    undefined_function(fname, orig_narg);
	}

	if(strcmp(funcall_table[addr].fname, fname) == 0 &&
	   funcall_table[addr].narg == narg)
	    break;

	addr = (addr + 1) % funcall_hash_size;	
    }

    return funcall_table[addr].f;
}

void make_funcall_table(void)
{
    int i;

    if(nmemfunc <= 0)
    {
	funcall_hash_size = 0;
	return;
    }

    funcall_hash_size = nmemfunc * 2;
    funcall_table = (funcall_table_t *)xmalloc(sizeof(funcall_table_t) * funcall_hash_size);

    for(i = 0; i < funcall_hash_size; i++)
	funcall_table[i].fname = NULL;	

    for(i = 0; i < nmemfunc; i++)
	install_funcall_table(i);
}

int obtain_funcall_address(QHProg* p, int* fi)
{
    MemFunction* f;
    int faddr;
    QHObj* name_obj;
    QHFunc x;

    name_obj = &memstack[MEMADDR(p->a1)];

    if(name_obj->type != QH_String)
	execute_error("None string arg 1 of `funcall'");

    faddr = search_funcall_table(name_obj->data.s->str, p->a2);
    f = memfunctab + faddr;

    x.func = faddr;
    x.pc   = qh_pc;
    x.ret  = MEMADDR(p->res);
    x.sp   = qh_sp;

    pushQHFuncStack(&fstk, x);

    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top + f->offsp, memstack_size);
    qh_argstart = qh_sp;
    qh_sp += f->offsp;
    qh_top = qh_sp;
    memstack[x.ret] = null_qhobj;
    *fi = faddr;

    return f->addr;
}

void qh_funcall(QHProg* p)
{
    MemFunction* f;
    int faddr;
    int narg = p->a2;
    int addr = obtain_funcall_address(p, &faddr);
    int ret  = MEMADDR(p->res);
    f = memfunctab + faddr;

    if(addr > 0)
    {
	qh_pc = addr - 1;
	if(f->narg < 0)
	    qh_vla_bind_proc(&p->a2);
    }
    else			/* call built in function */
    {
	if(f->extern_function)
	    qh_call_extern_function(f, narg, memstack + qh_argstart, memstack + ret);
	else
	    QH_CALL(-addr, (memstack + ret));
	qh_restore();
    }
}
