/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <limits.h>

#include "common.h"
#include "interp.h"
#include "functab.h"
#include "gutil.h"
#include "builtin.h"

#define MAX_SCREEN 33		/* number of screens */
#define DEFAULT_FONT "a14"
#define DEFAULT_JFONT "k14"

static XGCValues gc_values;	/* current gc values */
#define GC_LINE_WIDTH	(gc_values.line_width)
#define GC_CAP_STYLE	(gc_values.cap_style)
#define GC_JOIN_STYLE	(gc_values.join_style)
#define AllGCMask	(GCForeground | \
			 GCBackground | \
			 GCLineWidth  | \
			 GCLineStyle  | \
			 GCCapStyle   | \
			 GCJoinStyle)

static Display* disp = NULL;	/* display */
static Drawable win[MAX_SCREEN]; /* window */
static GC       gc = NULL;	/* graphics context */
static XFontStruct* fs1;	/* font struct for 1 byte character */
static XFontStruct* fs2;	/* font struct for 2 byte character */
static XFontStruct* default_fs1 = NULL; /* default font */
static XFontStruct* default_fs2 = NULL; /* default font */
static int      current_screen;
		/* current screen number: must be between 0 and 32 */
static int      current_visual_screen = 0;

static Bool     is_trans_cood = False;
		/* If it's True, translate a coordinate. Thus, calculate in
                   real number */
static int      win_width;	/* window width */
static int      win_height;	/* window height */
static int      win_fg;		/* current window foreground */
static int      win_bg;		/* window background */
static double   win_cx;		/* window left-bottom x-coordinate */
static double   win_cy;		/* window left-bottom y-coordinate */
static double   win_cwidth;	/* window coordinate width */
static double   win_cheight;	/* window coordinate height */
static int	win_prev_x;	/* previous drawn x-coordinate */
static int	win_prev_y;	/* previous drawn y-coordinate */
static int	backing_store_mode = True;
		/* use backing store if it is True */
static char*	win_title_name = "QuickHandy"; /* window title name */
static char*	win_icon_name = "QuickHandy"; /* icon title name */
static XPoint*	pbuf = NULL;	/* buffer for XPoint */
static int	npbuf = 0;	/* pbuf size */
static int	flush_mode = True; /* XFlush */
static unsigned long dash_line_pattern = 0xffffffff;/* dash line pattern bit */
static unsigned long current_event_mask = (KeyPressMask | ButtonPressMask);

#define NEVENT_TYPE_LIST (sizeof(event_type_list) / sizeof(int))
static int event_type_list[] = {
    KeyPress,
    KeyRelease,
    ButtonPress,
    ButtonRelease,
    MotionNotify,
};

#define CHECK_OPEN if(disp == NULL || win[current_visual_screen] == None) \
                       execute_error("Window is not opened")
#define FLUSH      if(!current_screen && flush_mode) XFlush(disp)
#define USE_SCREEN(sn)							\
if(sn < 0 || sn >= MAX_SCREEN)						\
    execute_error("Invalid screen number");				\
if(!win[sn])								\
{									\
    win[sn] = XCreatePixmap(disp, win[current_visual_screen],		\
		    win_width, win_height, DefaultDepth(disp, 0));	\
    XSetForeground(disp, gc, win_bg);					\
    XFillRectangle(disp, win[sn], gc, 0, 0, win_width, win_height);	\
    XSetForeground(disp, gc, win_fg);					\
}									\

#define OPEN_DISP_IF_CLOSE if(disp == NULL) \
                               disp = open_display(display_host_name);

#define SCR_MIN -20000
#define SCR_MAX  20000

static int within_screen(double x)
{
    if(x < SCR_MIN)
	return SCR_MIN;
    if(x > SCR_MAX)
	return SCR_MAX;
    return (int)x;
}

static int obj_within_screen(QHObj* x)
{
    if(x->type == QH_Int)
    {
	if(x->data.i < SCR_MIN)
	    return SCR_MIN;
	if(x->data.i > SCR_MAX)
	    return SCR_MAX;
	return x->data.i;
    }

    if(x->type == QH_Real)
    {
	if(x->data.r < SCR_MIN)
	    return SCR_MIN;
	if(x->data.r > SCR_MAX)
	    return SCR_MAX;
	return (int)x->data.r;
    }

    if(x->type == QH_Complex)
    {
	if(x->data.c.re < SCR_MIN)
	    return SCR_MIN;
	if(x->data.c.re > SCR_MAX)
	    return SCR_MAX;
	return (int)x->data.c.re;
    }

    execute_error("Invalid data type is specified");
    return 0;			/* dmy */
}

static Bool check_event_type(int event_type)
{
    int i;

    for(i = 0; i < NEVENT_TYPE_LIST; i++)
	if(event_type == event_type_list[i])
	    return True;
    return False;
}

static Display* open_display(const char* host_name)
{
    Display* disp;

    if((disp = XOpenDisplay(host_name)) == NULL)
    {
	fprintf(stderr, "Can't open display `%s'.\n",
		XDisplayName(host_name));
	qhexit(1);
    }

    /* initialize fonts */
    fs1 = default_fs1 = gload_query_font(disp, DEFAULT_FONT);
    fs2 = default_fs2 = gload_query_font(disp, DEFAULT_JFONT);

    return disp;
}

static void gopen_proc(int width, int height)
{
    int i;
    XSizeHints hints;
    XEvent e;
    XGCValues gv;

    /* check width and height */
    if(width <= 0)
	execute_error("Window width is negative or zero");
    if(height <= 0)
	execute_error("Window height is negative or zero");
    if(width >= SCR_MAX || height >= SCR_MAX)
	execute_error("Too large window size");

    win_width   = width;
    win_height  = height;
    win_cx      = 0.0;
    win_cy      = 0.0;
    win_cwidth  = (double)width;
    win_cheight = (double)height;
    win_prev_x  = 0;
    win_prev_y  = 0;

    memstack[Q_GWINDOW_PWIDTH].type    = QH_Int;
    memstack[Q_GWINDOW_PWIDTH].data.i  = width;
    memstack[Q_GWINDOW_PHEIGHT].type   = QH_Int;
    memstack[Q_GWINDOW_PHEIGHT].data.i = height;
    memstack[Q_GWINDOW_WIDTH].type     = QH_Real;
    memstack[Q_GWINDOW_WIDTH].data.r   = (double)width;
    memstack[Q_GWINDOW_HEIGHT].type    = QH_Real;
    memstack[Q_GWINDOW_HEIGHT].data.r  = (double)height;

    is_trans_cood = False;

    for(i = 0; i < MAX_SCREEN; i++)
	win[i] = None;

    /* create Window and GC */
    win[current_visual_screen] = XCreateSimpleWindow(disp,
				DefaultRootWindow(disp),
				0, 0, width, height, 0, 0, win_bg);

    gv.graphics_exposures = False;
    gc = XCreateGC(disp, win[current_visual_screen], GCGraphicsExposures, &gv);

    /* initialize */
    XSetForeground(disp, gc, win_fg);
    if(!XGetGCValues(disp, gc, AllGCMask, &gc_values))
	fprintf(stderr, "Warning: Can't get GC values.\n");

    if(dash_line_pattern == 0xffffffff)
    {
	XSetLineAttributes(disp, gc,
                           GC_LINE_WIDTH,
                           LineSolid,
                           GC_CAP_STYLE,
                           GC_JOIN_STYLE);
	XSetFunction(disp, gc, GXcopy);
    }
    else if(dash_line_pattern == 0)
    {
	XSetLineAttributes(disp, gc,
                           GC_LINE_WIDTH,
                           LineSolid,
                           GC_CAP_STYLE,
                           GC_JOIN_STYLE);
	XSetFunction(disp, gc, GXclear);
    }
    else
    {
	char dash_list[32];
	int n, nrot;

	gmake_bitpattern_dash(dash_line_pattern, &nrot, dash_list, &n);
	XSetLineAttributes(disp, gc,
			   GC_LINE_WIDTH,
			   LineOnOffDash,
			   GC_CAP_STYLE,
			   GC_JOIN_STYLE);
	XSetDashes(disp, gc, nrot, dash_list, n);
    }

    XStoreName(disp, win[current_visual_screen], win_title_name);
    XSetIconName(disp, win[current_visual_screen], win_icon_name);
    XSelectInput(disp, win[current_visual_screen],
		 ButtonPressMask | KeyPressMask | ExposureMask);

    hints.flags = PMinSize | PMaxSize;
    hints.min_width  = hints.max_width  = width;
    hints.min_height = hints.max_height = height;
    XSetNormalHints(disp, win[current_visual_screen], &hints);

    /* map window */
    XMapWindow(disp, win[current_visual_screen]);

    /* wait until window expose  */
    while(XMaskEvent(disp, ExposureMask, &e), e.type != Expose)
	;

    XSelectInput(disp, win[current_visual_screen], current_event_mask);
    if(backing_store_mode)
	set_backing_store(disp, win[current_visual_screen],
			  backing_store_mode);

    XSync(disp, False);
}

/* gopen(width, height, "fg", "bg"); */
void qh_gopen(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    int width, height;
    void qh_gclose(QHObj*);

    width  = obj_within_screen(a1);
    height = obj_within_screen(a2);

    /* close display if window is created */
    if(disp != NULL && win[current_visual_screen] != None)
	qh_gclose(NULL);

    /* open display */
    disp = open_display(display_host_name);

    /* obtain color pixel value */
    if(a3->type == QH_String)
	win_fg = MyColor(disp, cast_to_string(a3));
    else
	win_fg = cast_to_int(a3);

    if(a4->type == QH_String)
	win_bg = MyColor(disp, cast_to_string(a4));
    else
	win_bg = cast_to_int(a4);

    /* open the window */
    gopen_proc(width, height);
}

void qh_gclose(QHObj* res)
{
    int i;

    if(disp == NULL)
	return;			/* Do nothing. */

    /* free fonts */
    if(fs1 != default_fs1)
	XFreeFont(disp, fs1);
    if(fs2 != default_fs2)
	XFreeFont(disp, fs2);
    XFreeFont(disp, default_fs1);
    XFreeFont(disp, default_fs2);

    if(win[current_visual_screen] != None)
    {
	XFreeGC(disp, gc);
	XDestroyWindow(disp, win[current_visual_screen]);
	win[current_visual_screen] = None;
    }

    for(i = 0; i < MAX_SCREEN; i++)
	if(win[i] != None)
	{
	    XFreePixmap(disp, win[i]);
	    win[i] = None;
	}

    XCloseDisplay(disp);
    disp = NULL;
}

void qh_gwindow(QHObj* res)
{
    CHECK_OPEN;

    is_trans_cood = True;
    win_cx      = cast_to_real(STACK(1));
    win_cy      = cast_to_real(STACK(2));
    win_cwidth  = cast_to_real(STACK(3)) - win_cx;
    win_cheight = cast_to_real(STACK(4)) - win_cy;

    memstack[Q_GWINDOW_WIDTH].type    = QH_Real;
    memstack[Q_GWINDOW_WIDTH].data.r  = win_cwidth;
    memstack[Q_GWINDOW_HEIGHT].type   = QH_Real;
    memstack[Q_GWINDOW_HEIGHT].data.r = win_cheight;

    if(win_cwidth == 0.0 ||
       win_cheight == 0.0)
    {
	sprintf(temp_buff, "Bad window size: (%e, %e)-(%e, %e)",
		win_cx, win_cy, win_cwidth + win_cx, win_cheight + win_cy);
	execute_error(temp_buff);
    }
}

void qh_gscreen(QHObj* res)
{
    int scr;
    CHECK_OPEN;

    scr = cast_to_int(STACK(1));

    if(scr < 0 ||
       scr >= MAX_SCREEN)
    {
	sprintf(temp_buff, "Out of range screen number: %d", scr);
	execute_error(temp_buff);
    }

    USE_SCREEN(scr);
    current_screen = scr;
    memstack[Q_GCURRENT_SCREEN].type   = QH_Int;
    memstack[Q_GCURRENT_SCREEN].data.i = scr;
}

void qh_gcopy(QHObj* res)
{
    int scr1 = cast_to_int(STACK(1));
    int scr2 = cast_to_int(STACK(2));
    CHECK_OPEN;

    USE_SCREEN(scr1);
    USE_SCREEN(scr2);

    if(scr1 != scr2)
    {
	XCopyArea(disp, win[scr1], win[scr2], gc,
		  0, 0, win_width, win_height, 0, 0);
	if(!scr2 && flush_mode)
	    XFlush(disp);
    }
}

static void translate_coordinate(double rx, double ry, int* ix, int* iy)
{
    int x, y;

    x = within_screen((rx - win_cx) * (double)win_width / win_cwidth + 0.5);
    y = within_screen((ry - win_cy) * (double)win_height / win_cheight + 0.5);

    *ix = x;
    *iy = y;
}

static void reverse_coordinate(int ix, int iy, double* rx, double* ry)
{
    *rx = (double)ix * win_cwidth / (double)win_width + win_cx;
    *ry = (double)iy * win_cheight / (double)win_height + win_cy;
}

void qh_gcopy_area(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    QHObj* a5 = STACK(5);
    QHObj* a6 = STACK(6);
    QHObj* a7 = STACK(7);
    QHObj* a8 = STACK(8);
    int scr1;
    int scr2;
    int x, y, w, h, xx, yy;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a3), cast_to_real(a4), &x, &y);
	translate_coordinate(cast_to_real(a7), cast_to_real(a8), &xx, &yy);

	w = within_screen(cast_to_real(a5) * win_width / win_cwidth);
	h = within_screen(cast_to_real(a6) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a3);
	y = obj_within_screen(a4);
	xx = obj_within_screen(a7);
	yy = obj_within_screen(a8);
	w = obj_within_screen(a5);
	h = obj_within_screen(a6);
    }
/*
    y -= h;
    yy -= h;
*/
    scr1 = cast_to_int(a1);
    scr2 = cast_to_int(a2);

    USE_SCREEN(scr1);
    USE_SCREEN(scr2);

    XCopyArea(disp, win[scr1], win[scr2], gc, x, y, w, h, xx, yy);
    if(!scr2 && flush_mode)
	XFlush(disp);
}

void qh_gbacking_store(QHObj* res)
{
    int mode = cast_to_int(STACK(1));

    if(disp != NULL &&
       win[current_visual_screen] != None &&
       backing_store_mode != mode)
    {
	set_backing_store(disp, win[current_visual_screen], mode);
	XFlush(disp);
    }
    backing_store_mode = mode;
}

void qh_gsync(QHObj* res)
{
    CHECK_OPEN;
    XSync(disp, cast_to_int(STACK(1)));
}

void qh_gline(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    int x0, y0, x1, y1;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x0, &y0);
	translate_coordinate(cast_to_real(a3), cast_to_real(a4), &x1, &y1);
    }
    else
    {
	x0 = obj_within_screen(a1);
	y0 = obj_within_screen(a2);
	x1 = obj_within_screen(a3);
	y1 = obj_within_screen(a4);
    }
    XDrawLine(disp, win[current_screen], gc, x0, y0, x1, y1);
    win_prev_x = x1;
    win_prev_y = y1;

    FLUSH;
}

void qh_gline2(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    int x, y;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
    }
    XDrawLine(disp, win[current_screen], gc, win_prev_x, win_prev_y, x, y);
    win_prev_x = x;
    win_prev_y = y;

    FLUSH;
}

void qh_gpset(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    int x, y;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
    }
    XDrawPoint(disp, win[current_screen], gc, x, y);
    win_prev_x = x;
    win_prev_y = y;

    FLUSH;
}

void qh_gbox(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    int x, y, w, h;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);

	w = within_screen(cast_to_real(a3) * win_width / win_cwidth);
	h = within_screen(cast_to_real(a4) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
	w = obj_within_screen(a3);
	h = obj_within_screen(a4);
    }

    if(w < 0)
    {
	w = -w;
	x -= w;
    }

    if(h < 0)
    {
	h = -h;
	y -= h;
    }

    XDrawRectangle(disp, win[current_screen], gc, x, y, w, h);
    FLUSH;
}

void qh_gfbox(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    int x, y, w, h;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);

	w = (int)(cast_to_real(a3) * win_width / win_cwidth);
	h = (int)(cast_to_real(a4) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
	w = obj_within_screen(a3);
	h = obj_within_screen(a4);
    }

    if(w < 0)
    {
	w = -w;
	x -= w;
    }

    if(h < 0)
    {
	h = -h;
	y -= h;
    }

    XFillRectangle(disp, win[current_screen], gc, x, y, w, h);
    FLUSH;
}

static void array_point_converter(Array* a, XPoint* p, int n)
{
    int i;
    int x, y;
    union qhobj_template
    {
	void* p;
	int* i;
	double* r;
	XComplex* c;
    } u;

    switch(a->type)
    {
      case QH_Int:
      case QH_Real:
	if(a->dim != 2 ||
	   a->size[1] != 2)
	    execute_error("Wrong array dimension");
	break;
      case QH_Complex:
	if(a->dim != 1)
	    execute_error("Wrong array dimension");
	break;
    }

    u.p = a->elem;
    switch(a->type)
    {
      case QH_Int:
	if(is_trans_cood)
	{
	    for(i = 0; i < n; i++)
	    {
		translate_coordinate((double)u.i[2 * i],
				     (double)u.i[2 * i + 1],
				     &x, &y);
		pbuf[i].x = x;
		pbuf[i].y = y;
	    }
	}
	else
	{
	    for(i = 0; i < n; i++)
	    {
		pbuf[i].x = u.i[2 * i];
		pbuf[i].y = u.i[2 * i + 1];
	    }
	    x = pbuf[n - 1].x;
	    y = pbuf[n - 1].y;
	}
	break;
      case QH_Real:
	if(is_trans_cood)
	{
	    for(i = 0; i < n; i++)
	    {
		translate_coordinate(u.r[2 * i], u.r[2 * i + 1], &x, &y);
		pbuf[i].x = x;
		pbuf[i].y = y;
	    }
	}
	else
	{
	    for(i = 0; i < n; i++)
	    {
		pbuf[i].x = (int)u.r[2 * i];
		pbuf[i].y = (int)u.r[2 * i + 1];
	    }
	    x = pbuf[n - 1].x;
	    y = pbuf[n - 1].y;
	}
	break;
      case QH_Complex:
	if(is_trans_cood)
	{
	    for(i = 0; i < n; i++)
	    {
		translate_coordinate(u.c[i].re, u.c[i].im, &x, &y);
		pbuf[i].x = x;
		pbuf[i].y = y;
	    }
	}
	else
	{
	    for(i = 0; i < n; i++)
	    {
		pbuf[i].x = (int)u.c[i].re;
		pbuf[i].y = (int)u.c[i].im;
	    }
	    x = pbuf[n - 1].x;
	    y = pbuf[n - 1].y;
	}
	break;
    }

    win_prev_x = x;
    win_prev_y = y;
}

static void use_pbuf(int n)
{
    if(pbuf == NULL)
    {
	npbuf = n;
	pbuf = (XPoint *)xmalloc(sizeof(XPoint) * npbuf);
    }
    else if(npbuf < n)
    {
	npbuf = n;
	pbuf = (XPoint *)xrealloc(pbuf, sizeof(XPoint) * npbuf);
    }
}

void qh_gpoly(QHObj* res)
{
    QHObj* xo1 = STACK(1);
    QHObj* xo2 = STACK(2);
    Array* a;
    int n;
    CHECK_OPEN;

    if(xo1->type != QH_Array ||
       xo2->type == QH_Array)
	execute_error("Wrong gpoly argument");

    a = xo1->data.a;
    n = cast_to_int(xo2);

    if(n <= 1)
	return;

    use_pbuf(n);

    array_point_converter(a, pbuf, n);

    XDrawLines(disp, win[current_screen], gc,
	       pbuf, n, CoordModeOrigin);

    FLUSH;
}


void qh_gfpoly(QHObj* res)
{
    QHObj* xo1 = STACK(1);
    QHObj* xo2 = STACK(2);
    Array* a;
    int n;
    CHECK_OPEN;

    if(xo1->type != QH_Array ||
       xo2->type == QH_Array)
	execute_error("Wrong gfpoly argument");

    a = xo1->data.a;
    n = cast_to_int(xo2);

    if(n <= 1)
	return;

    use_pbuf(n);

    array_point_converter(a, pbuf, n);

    XFillPolygon(disp, win[current_screen], gc,
		 pbuf, n, Complex, CoordModeOrigin);

    FLUSH;
}

void qh_gcircle(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    int x, y, w, h;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
	w = within_screen(cast_to_real(a3) * win_width / win_cwidth);
	h = within_screen(cast_to_real(a3) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
	w = h = obj_within_screen(a3);
    }

    if(w < 0)
	w = -w;
    if(h < 0)
	h = -h;
    x -= w;
    y -= h;
    w *= 2;
    h *= 2;

    XDrawArc(disp, win[current_screen], gc, x, y, w, h, 0, 360 * 64);

    FLUSH;
}

void qh_gfcircle(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    int x, y, w, h;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
	w = within_screen(cast_to_real(a3) * win_width / win_cwidth);
	h = within_screen(cast_to_real(a3) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
	w = h = obj_within_screen(a3);
    }

    if(w < 0)
	w = -w;
    if(h < 0)
	h = -h;
    x -= w;
    y -= h;
    w *= 2;
    h *= 2;

    XFillArc(disp, win[current_screen], gc, x, y, w+1, h+1, 0, 360 * 64);
    FLUSH;
}

static inline int radian_arc(double th)
{
    return (int)(th / M_PI * (double)(180 * 64));
}

void qh_garc(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    QHObj* a5 = STACK(5);
    QHObj* a6 = STACK(6);
    int x, y, w, h, arg1, arg2, tmp;

    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
	w = within_screen(cast_to_real(a3) * win_width / win_cwidth);
	h = within_screen(cast_to_real(a4) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
	w = obj_within_screen(a3);
	h = obj_within_screen(a4);
    }

    arg1 = radian_arc(cast_to_real(a5));
    arg2 = radian_arc(cast_to_real(a6));

    if(w < 0)
    {
	w = -w;
	tmp = arg1;
	arg1 = 180 * 64 - arg2;
	arg2 = 180 * 64 - tmp;

    }
    if(h < 0)
    {
	h = -h;
	tmp = arg1;
	arg1 = 360 * 64 - arg2;
	arg2 = 360 * 64 - tmp;
    }
    x -= w;
    y -= h;
    w *= 2;
    h *= 2;

    XDrawArc(disp, win[current_screen], gc, x, y, w, h, arg1, arg2 - arg1);

    FLUSH;
}

void qh_gfarc(QHObj* res)
{
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    QHObj* a3 = STACK(3);
    QHObj* a4 = STACK(4);
    QHObj* a5 = STACK(5);
    QHObj* a6 = STACK(6);
    int x, y, w, h, arg1, arg2, tmp;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
	w = within_screen(cast_to_real(a3) * win_width / win_cwidth);
	h = within_screen(cast_to_real(a4) * win_height / win_cheight);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
	w = obj_within_screen(a3);
	h = obj_within_screen(a4);
    }

    arg1 = radian_arc(cast_to_real(a5));
    arg2 = radian_arc(cast_to_real(a6));

    if(w < 0)
    {
	w = -w;
	tmp = arg1;
	arg1 = 180 * 64 - arg2;
	arg2 = 180 * 64 - tmp;

    }
    if(h < 0)
    {
	h = -h;
	tmp = arg1;
	arg1 = 360 * 64 - arg2;
	arg2 = 360 * 64 - tmp;
    }
    x -= w;
    y -= h;
    w *= 2;
    h *= 2;

    XFillArc(disp, win[current_screen], gc, x-1, y, w+1, h+1,
	     arg1, arg2 - arg1);

    FLUSH;
}

void qh_gcolor_name(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    res->type = QH_Int;
    res->data.i = MyColor(disp, cast_to_string(STACK(1)));
}

void qh_gcolor_rgb(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    res->type = QH_Int;
    res->data.i = RGB(disp,
		      cast_to_int(STACK(1)),
		      cast_to_int(STACK(2)),
		      cast_to_int(STACK(3)));
}

void qh_gpen(QHObj* res)
{
    int fg;
    QHObj* color = STACK(1);
    CHECK_OPEN;

    if(IS_QHSTRING(color))
    {
	if(color->data.s->str == NULL)
	    fg = MyColor(disp, "");	/* WARNING */
	else
	    fg = MyColor(disp, (char *)color->data.s->str);
    }
    else
	fg = cast_to_int(color);

    if(fg != win_fg)
    {
	win_fg = fg;
	XSetForeground(disp, gc, fg);
	gc_values.foreground = fg;
    }
}

void qh_gpen3(QHObj* res)
{
    int r = cast_to_int(STACK(1));
    int g = cast_to_int(STACK(2));
    int b = cast_to_int(STACK(3));
    int fg;
    CHECK_OPEN;

    fg = RGB(disp, r, g, b);

    if(fg != win_fg)
    {
	win_fg = fg;
	XSetForeground(disp, gc, fg);
	gc_values.foreground = fg;
    }
}

static void set_pointer(int x, int y)
{
    if(is_trans_cood)
    {
	memstack[Q_GPOINTERX].type = QH_Real;
	memstack[Q_GPOINTERY].type = QH_Real;
	reverse_coordinate(x, y,
			   &memstack[Q_GPOINTERX].data.r,
			   &memstack[Q_GPOINTERY].data.r);
    }
    else
    {
	memstack[Q_GPOINTERX].type = QH_Int;
	memstack[Q_GPOINTERY].type = QH_Int;
	memstack[Q_GPOINTERX].data.i = x;
	memstack[Q_GPOINTERY].data.i = y;
    }
}

void qh_gpause(QHObj* res)
{
    XEvent e;
    CHECK_OPEN;

    do
    {
	XNextEvent(disp, &e);
    } while(check_event_type(e.type) == False);

    memstack[Q_GEVENT_TYPE].type = QH_Int;
    memstack[Q_GEVENT_TYPE].data.i = e.type;
    *res = memstack[Q_GEVENT_TYPE];

    memstack[Q_GEVENT_VALUE].type = QH_Int;
    switch(e.type)
    {
      case KeyPress:
      case KeyRelease:
	set_pointer(e.xkey.x, e.xkey.y);
	memstack[Q_GEVENT_VALUE].data.i = glookup_char(&e.xkey);
	break;
      case ButtonPress:
      case ButtonRelease:
	set_pointer(e.xbutton.x, e.xbutton.y);
	memstack[Q_GEVENT_VALUE].data.i = e.xbutton.button;
	break;
      case MotionNotify:
	set_pointer(e.xmotion.x, e.xmotion.y);
	memstack[Q_GEVENT_VALUE].data.i = 0; /* undefined */
	break;
      default:
	fprintf(stderr, "Systematic error: Unknown event type:%d\n", e.type);
	qhexit(1);
    }
}

void qh_gpause1(QHObj* res)
{
    double t = cast_to_real(STACK(1));
    XEvent e;
    Bool state;
    CHECK_OPEN;

    do
    {
	state = gsleep_event(disp, &e, t);
	if(state == False)
	{
	    memstack[Q_GEVENT_TYPE].type = QH_Int;
	    memstack[Q_GEVENT_TYPE].data.i = None;
	    *res = memstack[Q_GEVENT_TYPE];
	    return;
	}
    } while(check_event_type(e.type) == False);

    memstack[Q_GEVENT_TYPE].type = QH_Int;
    memstack[Q_GEVENT_TYPE].data.i = e.type;
    *res = memstack[Q_GEVENT_TYPE];

    memstack[Q_GEVENT_VALUE].type = QH_Int;
    switch(e.type)
    {
      case KeyPress:
      case KeyRelease:
	set_pointer(e.xkey.x, e.xkey.y);
	memstack[Q_GEVENT_VALUE].data.i = glookup_char(&e.xkey);
	break;
      case ButtonPress:
      case ButtonRelease:
	set_pointer(e.xbutton.x, e.xbutton.y);
	memstack[Q_GEVENT_VALUE].data.i = e.xbutton.button;
	break;
      case MotionNotify:
	set_pointer(e.xmotion.x, e.xmotion.y);
	memstack[Q_GEVENT_VALUE].data.i = 0; /* undefined */
	break;
      default:
	fprintf(stderr, "Systematic error: Unknown event type:%d\n", e.type);
	qhexit(1);
    }
}

void qh_gbutton(QHObj* res)
{
    XEvent e;
    CHECK_OPEN;

    res->type = QH_Int;
    if(XCheckTypedEvent(disp, ButtonPress, &e) == False)
    {
	res->data.i = None;
	return;
    }
    res->data.i = e.xbutton.button;
    set_pointer(e.xbutton.x, e.xbutton.y);
}

void qh_gkey(QHObj* res)
{
    XEvent e;
    CHECK_OPEN;

    res->type = QH_Int;
    if(XCheckTypedEvent(disp, KeyPress, &e) == False)
    {
	res->data.i = None;
	return;
    }
    res->data.i = glookup_char(&e.xkey);
    set_pointer(e.xbutton.x, e.xbutton.y);
}

void qh_gpointer(QHObj* res)
{
    Bool is_wait = cast_to_int(STACK(1));
    int x, y;
    CHECK_OPEN;

    if(!is_wait)
    {
	gget_pointer(disp, win[current_visual_screen], &x, &y);
    }
    else
    {
	XEvent e;
	unsigned long mask;

	mask = gget_event_mask(disp, win[current_visual_screen]);
	XSelectInput(disp, win[current_visual_screen],
		     mask | PointerMotionMask);
	XMaskEvent(disp, PointerMotionMask, &e);
	XSelectInput(disp, win[current_visual_screen],
		     mask & ~PointerMotionMask);

	x = e.xmotion.x;
	y = e.xmotion.y;
    }

    set_pointer(x, y);

    res->type = QH_Complex;
    reverse_coordinate(x, y, &res->data.c.re, &res->data.c.im);
}

void qh_gstring(QHObj* res)
{

    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    unsigned char* str = cast_to_string(STACK(3));
    int x, y;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
    }

    if(win_cheight < 0)
	y -= MAX(FONTHEIGHT(fs1), FONTHEIGHT(fs2));

    gmb_draw_string(disp, win[current_screen], gc, fs1, fs2, x, y, str);

    FLUSH;
}

void qh_gfont(QHObj* res)
{
    char* fname = cast_to_string(STACK(1));
    OPEN_DISP_IF_CLOSE;

    if(fs1 != default_fs1)
	XFreeFont(disp, fs1);
    fs1 = gload_query_font(disp, fname);
}

void qh_gjfont(QHObj* res)
{
    char* fname = cast_to_string(STACK(1));
    OPEN_DISP_IF_CLOSE;

    if(fs2 != default_fs2)
	XFreeFont(disp, fs2);
    fs2 = gload_query_font(disp, fname);
}

void qh_gspline(QHObj* res)
{
    QHObj* xo1 = STACK(1);
    QHObj* xo2 = STACK(2);
    Array* a;
    int n;
    CHECK_OPEN;

    if(xo1->type != QH_Array ||
       xo2->type == QH_Array)
	execute_error("Wrong gspline argument");

    a = xo1->data.a;
    n = cast_to_int(xo2);

    if(n <= 1)
	return;

    use_pbuf(n);

    array_point_converter(a, pbuf, n);

    draw_gspline(disp, win[current_screen], gc, pbuf, n);

    FLUSH;
}

void qh_gfspline(QHObj* res)
{
    QHObj* xo1 = STACK(1);
    QHObj* xo2 = STACK(2);
    Array* a;
    int n;
    CHECK_OPEN;

    if(xo1->type != QH_Array ||
       xo2->type == QH_Array)
	execute_error("Wrong gfspline argument");

    a = xo1->data.a;
    n = cast_to_int(xo2);

    if(n <= 1)
	return;

    use_pbuf(n);

    array_point_converter(a, pbuf, n);

    draw_gfspline(disp, win[current_screen], gc, pbuf, n);

    FLUSH;
}

void qh_gclear(QHObj* res)
{
    int scrno = cast_to_int(STACK(1));
    CHECK_OPEN;
    USE_SCREEN(scrno);

    if(scrno == 0)
    {
	XClearWindow(disp, win[current_visual_screen]);
	XFlush(disp);
    }
    else
    {
	XSetForeground(disp, gc, win_bg);
	XFillRectangle(disp, win[scrno], gc, 0, 0, win_width, win_height);
	XSetForeground(disp, gc, win_fg);
    }
}

void qh_gtitle_name(QHObj* res)
{
    win_title_name = cast_to_string(STACK(1));
    if(disp != NULL &&
       win[current_visual_screen] != None)
	XStoreName(disp, win[current_visual_screen], win_title_name);
}

void qh_gicon_name(QHObj* res)
{
    win_icon_name = cast_to_string(STACK(1));
    if(disp != NULL &&
       win[current_visual_screen] != None)
	XSetIconName(disp, win[current_visual_screen], win_icon_name);
}

void qh_gcolor_fg(QHObj* res)
{
    res->type = QH_Int;
    res->data.i = win_fg;
}

void qh_gflush(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    flush_mode = cast_to_int(STACK(1));
    FLUSH;
}

void qh_gtext_width(QHObj* res)
{
    int w;
    OPEN_DISP_IF_CLOSE;

    w = gmb_text_width(fs1, fs2, cast_to_string(STACK(1)));

    if(is_trans_cood)
    {
	res->type   = QH_Real;
	res->data.r = (double)w * win_width / (double)win_cwidth;
    }
    else
    {
	res->type   = QH_Int;
	res->data.i = w;
    }
}

void qh_gtext_height(QHObj* res)
{
    int h;
    OPEN_DISP_IF_CLOSE;

    h = MAX(FONTHEIGHT(fs1), FONTHEIGHT(fs2));

    if(is_trans_cood)
    {
	res->type   = QH_Real;
	res->data.r = (double)h * win_height / (double)win_cheight;
    }
    else
    {
	res->type   = QH_Int;
	res->data.i = h;
    }
}

void qh_gpoint(QHObj* res)
{
    int pixel;
    XImage* im;
    QHObj* a1 = STACK(1);
    QHObj* a2 = STACK(2);
    int x, y;
    CHECK_OPEN;

    if(is_trans_cood)
    {
	translate_coordinate(cast_to_real(a1), cast_to_real(a2), &x, &y);
    }
    else
    {
	x = obj_within_screen(a1);
	y = obj_within_screen(a2);
    }


    if(x < 0 ||
       x >= win_width ||
       y < 0 ||
       y >= win_height)
    {
	pixel = -1;
    }
    else
    {
	im = XGetImage(disp, win[current_visual_screen], x, y, 1, 1,
		       AllPlanes, ZPixmap);
	pixel = XGetPixel(im, 0, 0);
	XDestroyImage(im);
    }

    res->type = QH_Int;
    res->data.i = pixel;
}

void qh_gbell(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    XBell(disp, 100);
    XSync(disp, False);
}

void qh_glookup_color(QHObj* res)
{
    char* s;
    XColor c1, c;
    QHObj* arr_obj = STACK(2);
    Array* a;
    QHObj* color = STACK(1);
    union qhobj_template
    {
	void* p;
	int* i;
	double* r;
	QHObj* x;
    } u;
    OPEN_DISP_IF_CLOSE;

    if(arr_obj->type != QH_Array)
	execute_error("None array object");
    a = arr_obj->data.a;

    if(a->dim != 1)
	execute_error("None one dimensional array");

    if(a->nelem < 3)
	execute_error("Too short array size");

    if(IS_QHSTRING(color))
    {
	s = (char *)cast_to_string(color);
	if(!XLookupColor(disp, DefaultColormap(disp, 0), s, &c1, &c))
	    fprintf(stderr, "Warning: Can't lookup color name `%s'\n", s);
    }
    else
    {
	c.pixel = cast_to_int(color);
	XQueryColor(disp, DefaultColormap(disp, 0), &c);
    }

    u.p = a->elem;
    switch(a->type)
    {
      case QH_Int:
	u.i[0] = c.red / 257;
	u.i[1] = c.green / 257;
	u.i[2] = c.blue / 257;
	break;
      case QH_Real:
	u.r[0] = (double)c.red / 257;
	u.r[1] = (double)c.green / 257;
	u.r[2] = (double)c.blue / 257;
	break;
      case QH_Any:
	u.x[0].data.i = c.red / 257;
	u.x[1].data.i = c.green / 257;
	u.x[2].data.i = c.blue / 257;
	u.x[0].type = QH_Int;
	u.x[1].type = QH_Int;
	u.x[2].type = QH_Int;
	break;
      default:
	execute_error("Invalid array data type");
    }
}

void qh_gsetdash(QHObj* res)
{
    unsigned long x = (unsigned long)cast_to_int(STACK(1));
    char dash_list[32];
    int n, nrot;
    OPEN_DISP_IF_CLOSE;

    dash_line_pattern = x;
    if(win[current_visual_screen] == None) /* window is closed */
	return;

    if(x == 0xffffffff)
    {
	XSetLineAttributes(disp, gc,
                           GC_LINE_WIDTH,
                           LineSolid,
                           GC_CAP_STYLE,
                           GC_JOIN_STYLE);
	XSetFunction(disp, gc, GXcopy);
	return;
    }

    if(x == 0)
    {
	XSetLineAttributes(disp, gc,
                           GC_LINE_WIDTH,
                           LineSolid,
                           GC_CAP_STYLE,
                           GC_JOIN_STYLE);
	XSetFunction(disp, gc, GXclear);
	return;
    }

    gmake_bitpattern_dash(x, &nrot, dash_list, &n);
    XSetLineAttributes(disp, gc,
		       GC_LINE_WIDTH,
		       LineOnOffDash,
		       GC_CAP_STYLE,
		       GC_JOIN_STYLE);
    XSetDashes(disp, gc, nrot, dash_list, n);
}

void qh_gpixel_per_mmx(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    res->type   = QH_Real;
    res->data.r = (double)DisplayWidth(disp, 0) / DisplayWidthMM(disp, 0);
}

void qh_gpixel_per_mmy(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    res->type   = QH_Real;
    res->data.r = (double)DisplayHeight(disp, 0) / DisplayHeightMM(disp, 0);
}

void qh_gdisplay_width(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    res->type   = QH_Int;
    res->data.i = DisplayWidth(disp, 0);
}

void qh_gdisplay_height(QHObj* res)
{
    OPEN_DISP_IF_CLOSE;
    res->type   = QH_Int;
    res->data.i = DisplayHeight(disp, 0);
}

void qh_gresize_screen(QHObj* res)
{
    int scrno = cast_to_int(STACK(1));
    int width = cast_to_int(STACK(2));
    int height = cast_to_int(STACK(3));
    CHECK_OPEN;

    if(scrno < 0 || scrno >= MAX_SCREEN)
	execute_error("Invalid screen number");
    if(width <= 0 ||
       height <= 0)
	execute_error("Invalid screen size");

    if(width >= SCR_MAX || height >= SCR_MAX)
	execute_error("Too large screen size");

    if(scrno == current_visual_screen)
    {
	XResizeWindow(disp, win[current_visual_screen], width, height);
	win_width = width;
	win_height = height;
	XFlush(disp);
    }
    else
    {
	Pixmap pm;

	pm = XCreatePixmap(disp, win[current_visual_screen],
			   width, height, DefaultDepth(disp, 0));
	XSetForeground(disp, gc, win_bg);
	XFillRectangle(disp, pm, gc, 0, 0, width, height);
	XSetForeground(disp, gc, win_fg);

	if(win[scrno] != None)
	    XFreePixmap(disp, win[scrno]);
	win[scrno] = pm;
    }
}

void qh_gselect_input(QHObj* res)
{
    int mask = cast_to_int(STACK(1));
    unsigned long old_mask = current_event_mask;
    current_event_mask = (unsigned long)mask;
    if(disp != NULL)
	XSelectInput(disp, win[current_visual_screen], current_event_mask);

    res->type = QH_Int;
    res->data.i = (int)old_mask;
}

void qh_gvisualize_screen(QHObj* res)
{
    int newscr = cast_to_int(STACK(1));
    static Pixmap tmp_pm = None;
    static prev_w, prev_h;
    Drawable dtmp;
    USE_SCREEN(newscr);

    if(newscr == current_visual_screen)
	return;

    if(disp == NULL || win[current_visual_screen] == None)
    {
	current_visual_screen = newscr;
	return;
    }

    if(tmp_pm != None && (prev_w != win_width || prev_h != win_height))
    {
	XFreePixmap(disp, tmp_pm);
	tmp_pm = None;
    }

    if(tmp_pm == None)
    {
	tmp_pm = XCreatePixmap(disp, win[current_visual_screen],
			       win_width, win_height,
			       DefaultDepth(disp, DefaultScreen(disp)));
	prev_w = win_width;
	prev_h = win_height;
    }

    XCopyArea(disp, win[current_visual_screen], tmp_pm, gc,
	      0, 0, win_width, win_height, 0, 0);
    XCopyArea(disp, win[newscr], win[current_visual_screen], gc,
	      0, 0, win_width, win_height, 0, 0);
    XCopyArea(disp, tmp_pm, win[newscr], gc,
	      0, 0, win_width, win_height, 0, 0);

    dtmp = win[newscr];
    win[newscr] = win[current_visual_screen];
    win[current_visual_screen] = dtmp;

    memstack[Q_GVISUAL_SCREEN].type = QH_Int;
    memstack[Q_GVISUAL_SCREEN].data.i = newscr;

    current_visual_screen = newscr;
    XFlush(disp);
}
