/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include "common.h"
#include "opstr.h"
#include "interp.h"
#include "qhc.h"
#include "qhfile.h"
#include "cpp.defs"
#ifdef USE_DLD_LIBRARY
#include <dld.h>
#endif


#define QH_QHL QHLIBDIR "/qh.qhl"

int nerror = 0;
char temp_buff[BUFSIZ];		/* for using temporary string */
int prog_argc;
char** prog_argv;
char_ptrStack cpp_arguments;
extern void compile_all(void);

/* argument flags */
Bool compile_flag = False;
Bool trace_mode = False;
Bool print_code_mode = False;
Bool print_garbage_collection = False;
Bool load_default_qh_library = True;
static Bool only_cpreprocessor = False;
Bool expr_optimazation = False;
/* Bool expr_optimazation = True;	 To debug optimizer */
Bool Wall_flag = False;
Bool W_uninitialized_variable = False;
Bool W_undeclared_global_variable = False;
Bool W_case_break = False;
Bool W_unused_local_variable;
char* display_host_name = NULL;
Bool qh_prof_mode = False;
Bool verbose_mode = False;

void qhexit(int status)
{
    qhfileclose();
    end_qhfile();

/* Ignore free_cell_object() after ver. 2.75
   free_cell_object(); */

    exit(status);
}

static void print_cpreprocessor_output(void)
{
    extern FILE* yyin;

    while(fgets(temp_buff, BUFSIZ, yyin))
	fputs(temp_buff, stdout);
    qhfileclose();
}

static Bool file_exist(const char* path)
{
    int fd;

    if((fd = open(path, 0)) < 0)
	return False;
    close(fd);
    return True;
}

void cat_file(const char* s)
{
    int fd;

    if((fd = open(s, 0)) < 0)
    {
	perror(s);
	exit(1);
    }

    while(read(fd, temp_buff, BUFSIZ) > 0)
	write(1, temp_buff, BUFSIZ);
    close(fd);
}

void usage(void)
{
    puts("qh [options] [filename] [usr-program-options]");
    puts("qh + [usr-program-options]");
    puts("options:");
    puts("\t-h    \tprint this message");
    puts("\t-trace    \ttrace mode");
    puts("\t-key [keyword ... ]    \tdisplay help of specified keywords");
    puts("\t-v    \tverbose mode");
    puts("\t-S    \tdisplay compiled internal code");
    puts("\t-Pgarbage    \tverbose garbage collection mode");
    puts("\t-I<dir>    \tinclude directory path");
    puts("\t-D<sym>    \tdefine for cpp");
    puts("\t-E    \tcpp only");
    puts("\t-O    \toptimize internal code, but start up speed is down");
    puts("\t-q    \tdon't include default library");
    puts("\t-Pglobal    \tdisplay global symbols");
    puts("\t-Wglobal    \twarning flags for undeclared global variable");
    puts("\t-Wall    \tset all warning flags exept for -Wglobal");
    puts("\t-Wuninitialized-variable    \twarning flags for uninitialized variable");
    puts("\t-Wunused-local-variable    \twarning flags for unused local variable");
    puts("\t-Wcase-break    \t(Not work this option)");
    puts("\t-disp <display-name>    \tspecified display name");
    puts("\t-include <include-filename>    \tspecified include filename");
    puts("\t-cpp <cpp-pathname>    \tspecified path of cpp");
    puts("\t-p    \tprof mode of internal code");
#ifdef USE_DLD_LIBRARY
    puts("\t-l <standard-unix-object-file(*.o)>    \tspecified compile external defined function");
#endif
    puts("\t-print-libdir    \tdisplay QuickHandy's library path");
    puts("\t-print-qhcflags    \tdisplay qhc compile flags");
    puts("\t-print-qhldflags    \tdisplay qhld compile flags");
}

void free_memory_in_compiling_at_once(void)
{
    static int flag = False;

    if(flag == False)
    {
	flag = True;
	delete_all_symbol();
	free_reuse_tree();
	free_temp();
	free_funcquad();
	free_quad_node();
	free_stmt();
	free_functab();
	delete_hash_table();
    }
}

char_ptrStack include_file_list;
void main(int argc, char** argv)
{
    int i;
    int Pglobal = False;
    char* qh_program_name;
    char* cpp_path = NULL;
#ifdef USE_DLD_LIBRARY
    char_ptrStack stdobj_list;
#endif
    extern int yyparse(void);
    extern void init_qhparse(void);
    void keyword_helper(int c, char** v);

    qh_program_name = NULL;
    makechar_ptrStack(&cpp_arguments, 16);
    makechar_ptrStack(&include_file_list, 16);
#ifdef USE_DLD_LIBRARY
    makechar_ptrStack(&stdobj_list, 16);
#endif

    /* command line option parsing */
    /*
      qh 
      qh -[opt]
      qh + [prog-opt];
      qh -[opt] + [prog-opt];
      qh filename
      qh -[opt] filename
      qh filename [prog-opt];
      qh -[opt] filename [prog-opt];
     */

    pushchar_ptrStack(&cpp_arguments, CPP);
    pushchar_ptrStack(&include_file_list, QH_QHL);

#define CHECK_ARG if(i == argc) {usage(); exit(1);}
    for(i = 1; i < argc; i++)
    {
	if(argv[i][0] == '-')
	{
	    if(argv[i][0] == '-' &&
	       (argv[i][1] == 'I' ||
		argv[i][1] == 'D')) /* -I or -D */
	    {
		if(argv[i][2] == '\0')
		{
		    i++;
		    CHECK_ARG;
		    pushchar_ptrStack(&cpp_arguments, argv[i - 1]);
		    pushchar_ptrStack(&cpp_arguments, argv[i]);
		}
		else
		    pushchar_ptrStack(&cpp_arguments, argv[i]);
	    }
	    else if(!strcmp(argv[i], "-h"))
	    {
		usage();
		qhexit(0);
	    }
	    else if(!strcmp(argv[i], "-trace"))
	    {
		trace_mode = True;
	    }
	    else if(!strcmp(argv[i], "-key"))
	    {
		keyword_helper(argc - i - 1, argv + i + 1);
		qhexit(0);
	    }
	    else if(!strcmp(argv[i], "-v"))
	    {
		if(!verbose_mode)
		    fprintf(stderr, "qh: Version %s\n", VERSION);
		verbose_mode = True;
	    }
	    else if(!strcmp(argv[i], "-S"))
	    {
		print_code_mode = True;
	    }
	    else if(!strcmp(argv[i], "-Pgarbage"))
	    {
		print_garbage_collection = True;
	    }
	    else if(!strcmp(argv[i], "-E"))
	    {
		only_cpreprocessor = True;
	    }
	    else if(!strcmp(argv[i], "-O"))
	    {
		expr_optimazation = True;
	    }
	    else if(!strcmp(argv[i], "-q"))
	    {
		load_default_qh_library = False;
	    }
	    else if(!strcmp(argv[i], "-Pglobal"))
	    {
		Pglobal = True;
	    }
	    else if(!strcmp(argv[i], "-Wall"))
	    {
		Wall_flag = True;
		W_uninitialized_variable = True;
		W_case_break = True;
		W_unused_local_variable = True;
	    }
	    else if(!strcmp(argv[i], "-Wunused-local-variable"))
	    {
		W_unused_local_variable = True;
	    }
	    else if(!strcmp(argv[i], "-Wuninitialized-variable"))
	    {
		W_uninitialized_variable = True;
	    }
	    else if(!strcmp(argv[i], "-Wglobal"))
	    {
		W_undeclared_global_variable = True;
	    }
	    else if(!strcmp(argv[i], "-Wcase-break"))
	    {
		W_case_break = True;
	    }
	    else if(!strcmp(argv[i], "-c"))
	    {
		compile_flag = True;
		print_code_mode = True;
	    }
	    else if(!strcmp(argv[i], "-disp"))
	    {
		i++;
		CHECK_ARG;
		display_host_name = argv[i];
	    }
	    else if(!strcmp(argv[i], "-cpp"))
	    {
		i++;
		CHECK_ARG;
		cpp_path = argv[i];
	    }
	    else if(!strcmp(argv[i], "-include"))
	    {
		i++;
		CHECK_ARG;
		pushchar_ptrStack(&include_file_list, argv[i]);
	    }
#ifdef USE_DLD_LIBRARY
	    else if(!strcmp(argv[i], "-l"))
	    {
		i++;
		CHECK_ARG;
		pushchar_ptrStack(&stdobj_list, argv[i]);
	    }
#endif
	    else if(!strcmp(argv[i], "-p"))
	    {
		qh_prof_mode = True;
	    }
	    else if(!strcmp(argv[i], "-print-libdir"))
	    {
		printf("%s\n", QHLIBDIR);
		qhexit(0);
	    }
	    else if(!strcmp(argv[i], "-print-qhcflags"))
	    {
		sprintf(temp_buff, "%s/qh-compile-cflags", QHLIBDIR);
		cat_file(temp_buff);
		qhexit(0);
	    }
	    else if(!strcmp(argv[i], "-print-qhldflags"))
	    {
		sprintf(temp_buff, "%s/qh-compile-ldflags", QHLIBDIR);
		cat_file(temp_buff);
		qhexit(0);
	    }
	    else
	    {
		usage();
		qhexit(1);
	    }
	    continue;
	}

	if(strcmp(argv[i], "+") != 0)
	    qh_program_name = argv[i];
	break;
    }
#undef CHECK_ARG
    prog_argc = argc - i - 1;
    prog_argv = argv + i + 1;

#ifdef USE_DLD_LIBRARY
    if(stdobj_list.ptr > 0)
    {
	if(file_exist(argv[0]))
	    dld_init(argv[0]);
	else
	{
	    strcpy(temp_buff, QHPATH);
	    dld_init(temp_buff);
	}

	for(i = 0; i < stdobj_list.ptr; i++)
	{
	    if(verbose_mode)
		fprintf(stderr, "loading: %s\n", stdobj_list.elem[i]);
	    dld_link(stdobj_list.elem[i]);
	}
    }
    deletechar_ptrStack(&stdobj_list);
#endif

    if(verbose_mode && qh_program_name == NULL)
	qhexit(0);

    if(cpp_path == NULL)
    {
#if 0
	pushchar_ptrStack(&cpp_arguments, "-lang-c++");
	if(load_default_qh_library)
	{
	    pushchar_ptrStack(&cpp_arguments, "-include");
	    pushchar_ptrStack(&cpp_arguments, QH_QHL);
	}
#endif
    }
    else
	cpp_arguments.elem[0] = cpp_path;
#if 0
    for(i = 0; i < include_file_list.ptr; i++)
    {
	pushchar_ptrStack(&cpp_arguments, "-include");
	pushchar_ptrStack(&cpp_arguments, include_file_list.elem[i]);
    }
#endif

    if(compile_flag)
    {
	puts("/*");
	fflush(stdout);
    }

    init_hash_table();
    init_temp();
    init_cell_object();
    init_symbol();
    init_functab();
    init_stmt();
    init_interp();
    init_qhparse();

    pushchar_ptrStack(&cpp_arguments, "-I" QHLIBDIR);
#ifdef INCROOT
    pushchar_ptrStack(&cpp_arguments, "-I" INCROOT);
#endif
    qhfileopen(qh_program_name);

    if(only_cpreprocessor)
    {
	print_cpreprocessor_output();
	qhexit(1);
    }
    else
    {
	yyparse();
    }

    qhfileclose();
    compile_all();
    deletechar_ptrStack(&cpp_arguments);

    if(Pglobal)
    {
	print_all_global_symbol();
	qhexit(0);
    }

    if(print_code_mode)
	print_qhprog();

    if(compile_flag)
	puts("*/");

    if(nerror == 0)
    {
	qh_pc = 1;
	qh_top = qh_sp;

	if(compile_flag)
	    compile_to_C(stdout, qhprog, get_current_qhprog());
	else
	{
	    if(print_code_mode == False &&
	       Pglobal == False)
	    {
#ifdef DEBUG
		free_memory_in_compiling_at_once();
#endif
		init_qhfile();
		interp(qhprog);
		end_qhfile();
	    }
	}
    }
    else
    {
	fprintf(stderr, "%d errors\n", nerror);
    }

    qhexit(!!nerror);
}
