/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "functab.h"
#include "interp.h"
#include "fdrand48.h"

#define SIGN(x) ((x) < 0 ? -1 : 1)
#ifndef _ABS
#define	_ABS(x) ((x) < 0 ? -(x) : (x))
#endif

static inline void sin_and_cos(double x, double* s, double* c)
{
#if defined(sun) && !defined(__svr4__) /* sun4 */
    sincos(x, s, c);
#else
    *s = sin(x);
    *c = cos(x);
#endif
}

/********************************* SIN **************************************/
void qh_sin(QHProg* q)
{
    QHObj* x   = memstack + MEMADDR((q)->a1);
    QHObj* res = memstack + MEMADDR((q)->res);
    double re, im;

    if(x->type >= QH_String)
	invalid_operation(q->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = sin((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = sin(x->data.r);
	break;

      case QH_Complex:
	re = x->data.c.re;
	im = x->data.c.im;
	res->type = QH_Complex;
	res->data.c.re = sin(re) * cosh(im);
	res->data.c.im = cos(re) * sinh(im);
	break;
    }
}

/********************************* COS **************************************/
void qh_cos(QHProg* q)
{
    QHObj* x   = memstack + MEMADDR((q)->a1);
    QHObj* res = memstack + MEMADDR((q)->res);
    double re, im;

    if(x->type >= QH_String)
	invalid_operation(q->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = cos((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = cos(x->data.r);
	break;

      case QH_Complex:
	re = x->data.c.re;
	im = x->data.c.im;
	res->type = QH_Complex;
	res->data.c.re =  cos(re) * cosh(im);
	res->data.c.im = -sin(re) * sinh(im);
	break;
    }
}

/********************************* TAN **************************************/
void qh_tan(QHProg* q)
{
    QHObj* x   = memstack + MEMADDR((q)->a1);
    QHObj* res = memstack + MEMADDR((q)->res);
    double re, im, t;

    if(x->type >= QH_String)
	invalid_operation(q->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = tan((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = tan(x->data.r);
	break;

      case QH_Complex:
	re = x->data.c.re * 2.0;
	im = x->data.c.im * 2.0;
	t  = cos(re) + cosh(im);
	res->type = QH_Complex;
	res->data.c.re = sin(re)  / t;
	res->data.c.im = sinh(im) / t;
	break;
    }
}

/********************************* ATAN2 *************************************/
void qh_atan2(QHProg* q)
{
    QHObj* x1  = memstack + MEMADDR((q)->a1);
    QHObj* x2  = memstack + MEMADDR((q)->a2);
    QHObj* res = memstack + MEMADDR((q)->res);
    double x, y;

    if(x1->type >= QH_Complex ||
       x2->type >= QH_Complex)
	invalid_operation(q->op);

    if(x1->type == QH_Real)
	y = x1->data.r;
    else
	y = (double)x1->data.i;

    if(x2->type == QH_Real)
	x = x2->data.r;
    else
	x = (double)x2->data.i;

    res->type   = QH_Real;
    res->data.r = atan2(y, x);
}

/********************************* ASIN **************************************/
void qh_asin(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = asin((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = asin(x->data.r);
	break;

      case QH_Complex:
	{
	    double t1, t2, re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    t1 = sqrt((1 + re) * (1 + re) + im * im);
	    t2 = sqrt((1 - re) * (1 - re) + im * im);
	    res->type = QH_Complex;
	    res->data.c.re =            asin ((t1 - t2) / 2);
	    res->data.c.im = SIGN(im) * acosh((t1 + t2) / 2);
	}
	break;
    }
}

/********************************* ACOS **************************************/
void qh_acos(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = acos((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = acos(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, t1, t2;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    t1 = sqrt((1 + re) * (1 + re) + im * im);
	    t2 = sqrt((1 - re) * (1 + re) + im * im);
	    res->type = QH_Complex;
	    res->data.c.re =            acos ((t1 - t2) * 0.5);
	    res->data.c.im = SIGN(im) * acosh((t1 + t2) * 0.5);
	}
	break;
    }
}

/********************************* ATAN **************************************/
void qh_atan(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = atan((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = atan(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, r;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    r = re * re + im * im;
	    res->type = QH_Complex;
	    res->data.c.re = atan (2 * re / (1 - r)) * 0.5;
	    res->data.c.im = atanh(2 * im / (1 + r)) * 0.5;
	}
	break;
    }
}

/********************************* SINH **************************************/
void qh_sinh(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = sinh((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = sinh(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, s, c;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    sin_and_cos(im, &s, &c);
	    res->data.c.re = sinh(re) * c;
	    res->data.c.im = cosh(re) * s;
	}
	break;
    }
}

/********************************* COSH **************************************/
void qh_cosh(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = cosh((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = cosh(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, s, c;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    sin_and_cos(im, &s, &c);
	    res->data.c.re = cosh(re) * c;
	    res->data.c.im = sinh(re) * s;
	}
	break;
    }
}

/********************************* TANH **************************************/
void qh_tanh(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = tanh((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = tanh(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, t;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    t = cosh(re) + cos(im);
	    res->type = QH_Complex;
	    res->data.c.re = sinh(re) / t;
	    res->data.c.im = sin (im) / t;
	}
	break;
    }
}

/********************************* ASINH **************************************/
void qh_asinh(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = asinh((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = asinh(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, t1, t2;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    t1 = sqrt(re * re + (1 + im) * (1 + im));
	    t2 = sqrt(re * re + (1 + im) * (1 - im));
	    res->type = QH_Complex;
	    res->data.c.re =            acosh((t1 + t2) * 0.5);
	    res->data.c.im = SIGN(im) * asin ((t1 - t2) * 0.5);
	}
	break;
    }
}

/********************************* ACOSH **************************************/
void qh_acosh(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = acosh((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = acosh(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, t1, t2;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    t1 = sqrt((1 + re) * (1 + re) + im * im);
	    t2 = sqrt((1 - re) * (1 + re) + im * im);
	    res->type = QH_Complex;
	    res->data.c.re =            acosh((t1 + t2) * 0.5);
	    res->data.c.im = SIGN(im) * acos ((t1 - t2) * 0.5);
	}
	break;
    }
}

/********************************* ATANH **************************************/
void qh_atanh(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = atanh((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = atanh(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, re, r;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    r  = re * re + im * im;
	    res->type = QH_Complex;
	    res->data.c.re = atanh(2 * re / (1 + r)) * 0.5;
	    res->data.c.im = atan (2 * im / (1 - r)) * 0.5;
	}
	break;
    }
}

/********************************* RAND2 **************************************/
void qh_rand2(QHProg* q)
{
    QHObj* x1  = memstack + MEMADDR((q)->a1);
    QHObj* x2  = memstack + MEMADDR((q)->a2);
    QHObj* res = memstack + MEMADDR((q)->res);
    double min, max;
    double imin, imax;

    if(x1->type >= QH_String ||
       x2->type >= QH_String)
	invalid_operation(q->op);

    if(x1->type <= QH_Real &&
       x2->type <= QH_Real)
    {
	min = cast_to_real(x1);
	max = cast_to_real(x2);
	res->type = QH_Real;
	res->data.r = min + fdrand48() * (max - min);
    }
    else
    {
	if(x1->type == QH_Complex)
	{
	    min  = x1->data.c.re;
	    imin = x1->data.c.im;
	}
	else
	{
	    min  = cast_to_real(x1);
	    imin = 0.0;
	}

	if(x2->type == QH_Complex)
	{
	    max  = x2->data.c.re;
	    imax = x2->data.c.im;
	}
	else
	{
	    max  = cast_to_real(x2);
	    imax = 0.0;
	}

	res->type = QH_Complex;
	res->data.c.re = min  + fdrand48() * (max  - min);
	res->data.c.im = imin + fdrand48() * (imax - imin);
    }
}

/********************************* RAND1 **************************************/
void qh_rand1(QHProg* q)
{
    QHObj* x   = memstack + MEMADDR((q)->a1);
    QHObj* res = memstack + MEMADDR((q)->res);

    if(x->type >= QH_String)
	invalid_operation(q->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = (double)x->data.i * fdrand48();
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = x->data.r * fdrand48();
	break;

      case QH_Complex:
	res->type = QH_Complex;
	res->data.c.re = x->data.c.re * fdrand48();
	res->data.c.im = x->data.c.im * fdrand48();
	break;
    }
}

/********************************* RAND0 **************************************/
void qh_rand0(QHProg* q)
{
    QHObj* res = memstack + MEMADDR((q)->res);

    res->type = QH_Real;
    res->data.r = fdrand48();
}

/********************************* SQRT **************************************/
void qh_sqrt(QHProg* q)
{
    QHObj* x   = memstack + MEMADDR((q)->a1);
    QHObj* res = memstack + MEMADDR((q)->res);
    double re, im, r, th;

    if(x->type >= QH_String)
	invalid_operation(q->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = sqrt((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = sqrt(x->data.r);
	break;

      case QH_Complex:
	re = x->data.c.re;
	im = x->data.c.im;

	res->type = QH_Complex;
	if(re == 0.0 &&
	   im == 0.0)
	{
	    res->data.c.re = 0.0;
	    res->data.c.im = 0.0;
	}
	else
	{
	    double s, c;

	    r = pow(re * re + im * im, 0.25);
	    th = atan2(im, re) * 0.5;
	    sin_and_cos(th, &s, &c);
	    res->data.c.re = r * c;
	    res->data.c.im = r * s;
	}
	break;
    }
}

/********************************* ABS **************************************/
void qh_abs(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    switch(x->type)
    {
      case QH_Int:
	{
	    int i;
	    res->type = QH_Int;
	    i = x->data.i;
	    res->data.i = _ABS(i);
	}
	return;

      case QH_Real:
	{
	    double r;
	    res->type = QH_Real;
	    r = x->data.r;
	    res->data.r = _ABS(r);
	}
	return;

      case QH_Complex:
	{
	    double re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Real;
	    res->data.r = sqrt(re * re + im * im);
	}
	return;
    }

    invalid_operation(p->op);
}

/********************************* FLOOR *************************************/
void qh_floor(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = x->data.i;
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = floor(x->data.r);
	break;

      case QH_Complex:
	{
	    double re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    res->data.c.re = floor(re);
	    res->data.c.im = floor(im);
	}
	break;
    }
}

/********************************* CEIL **************************************/
void qh_ceil(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = x->data.i;
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = ceil(x->data.r);
	break;

      case QH_Complex:
	{
	    double re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    res->data.c.re = ceil(re);
	    res->data.c.im = ceil(im);
	}
	break;
    }
}

/********************************* EXP **************************************/
void qh_exp(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = exp((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = exp(x->data.r);
	break;

      case QH_Complex:
	{
	    double im, r, s, c;

	    r = exp(x->data.c.re);
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    sin_and_cos(im, &s, &c);
	    res->data.c.re = r * c;
	    res->data.c.im = r * s;
	}
	break;
    }
}

/********************************* LOG **************************************/
void qh_log(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = log((double)x->data.i);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = log(x->data.r);
	break;

      case QH_Complex:
	{
	    double re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    res->data.c.re = log(sqrt(re * re + im * im));
	    res->data.c.im = atan2(im, re);
	}
	break;
    }
}

/******************************** LOG10 **************************************/
void qh_log10(QHProg* p)
{

    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = log((double)x->data.i) * (1.0 / M_LN10);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = log(x->data.r) * (1.0 / M_LN10);
	break;

      case QH_Complex:
	{
	    double re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    res->data.c.re = log(sqrt(re * re + im * im)) * (1.0 / M_LN10);
	    res->data.c.im = atan2(im, re) * (1.0 / M_LN10);
	}
	break;
    }
}

/********************************* LOG2 **************************************/
void qh_log2(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);

    if(x->type >= QH_String)
	invalid_operation(p->op);

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Real;
	res->data.r = log((double)x->data.i) * (1.0 / M_LN2);
	break;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = log(x->data.r) * (1.0 / M_LN2);
	break;

      case QH_Complex:
	{
	    double re, im;

	    re = x->data.c.re;
	    im = x->data.c.im;
	    res->type = QH_Complex;
	    res->data.c.re = log(sqrt(re * re + im * im)) * (1.0 / M_LN2);
	    res->data.c.im = atan2(im, re) * (1.0 / M_LN2);
	}
	break;
    }
}

/****************************** CONJUGATE ***********************************/
void qh_conjugate(QHProg* p)
{

    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);
    int type = x->type;

    if(type >= QH_String)
	invalid_operation(p->op);

    res->type = QH_Complex;
    switch(type)
    {
      case QH_Int:
	res->data.c.re = (double)x->data.i;
	res->data.c.im = 0.0;
	break;

      case QH_Real:
	res->data.c.re = x->data.r;
	res->data.c.im = 0.0;
	break;

      case QH_Complex:
	res->data.c.re = x->data.c.re;
	res->data.c.im = -x->data.c.im;
	break;
    }
}

/********************************* ARG **************************************/
void qh_arg(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);
    int type = x->type;

    if(type >= QH_String)
	invalid_operation(p->op);

    res->type = QH_Real;
    switch(type)
    {
      case QH_Int:
      case QH_Real:
	res->data.r = 0.0;
	break;

      case QH_Complex:
	if(x->data.c.re == 0.0 &&
	   x->data.c.im == 0.0)
	    res->data.r = 0.0;
	else
	    res->data.r = atan2(x->data.c.im, x->data.c.re);
	break;
    }
}
