/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "optmz_tree.h"

#define DOUBLE_EXPRESS "%.19e"

static Tree* integer_constant_folding2(Tree* t)
{
    int op;
    int left, right;
    double dval = 0;
    int ival = 0;
    int type = 0;

    op = t->val;
    left = atoi(STRING(t->left->val));
    right = atoi(STRING(t->right->val));
    reuse_tree(t->left), t->left = NULL;
    reuse_tree(t->right), t->right = NULL;

#define CONST_FOLDING(s, op) case s: ival = (left op right); type = T_CINT; break;
    switch(op)
    {
	CONST_FOLDING(ADD, +);
	CONST_FOLDING(SUB, -);
	CONST_FOLDING(MUL, *);
	CONST_FOLDING(DIV, /);
	CONST_FOLDING(REM, %);

      case POW:
	if(right == 0)
	    ival = 1;
	else if(left == 0 || right < 0)
	    ival = 0;
	else
	{
	    int i;
	    ival = left;
	    for(i = 1; i < right; i++)
		ival *= left;
	}
	type = T_CINT;
	break;

	CONST_FOLDING(OR, ||);
	CONST_FOLDING(AND, &&);
	CONST_FOLDING(EQ, ==);
	CONST_FOLDING(NQ, !=);
	CONST_FOLDING(LE, <=);
	CONST_FOLDING(LT, <);
	CONST_FOLDING(GE, >=);
	CONST_FOLDING(GT, >);
	CONST_FOLDING(ALSHIFT, <<);
	CONST_FOLDING(ARSHIFT, >>);

      case LLSHIFT:
	ival = (unsigned int)left << right;
	type = T_CINT;
	break;

      case LRSHIFT:
	ival = (unsigned int)left >> right;
	type = T_CINT;
	break;

      case ATAN2:
	dval = atan2((double)left, (double)right);
	type = T_CREAL;
	break;
 
      default:
	fprintf(stderr,	"Systematic error: Undefined bin-op: `%d', in integer_constant_folding2().\n",
		op);
	qhexit(1);
    }
#undef CONST_FOLDING

    if(type == T_CINT)
	sprintf(temp_buff, "%d", ival);
    else
	sprintf(temp_buff, DOUBLE_EXPRESS, dval);

    t->type = type;
    t->val  = install_string(temp_buff);

    return t;
}

static Tree* integer_constant_folding1(Tree* t)
{
    int op;
    int left;
    int type = 0;
    int    ival = 0;
    double dval = 0;

    op = t->val;
    left = atoi(STRING(t->left->val));
    reuse_tree(t->left), t->left = NULL;

#define ICONST_FOLDING(s, op) case s: type = T_CINT; ival = (op left); break;
#define IMATH_FUNC_FOLDING(s, func) case s: type = T_CREAL; dval = func((double)left); break;

    switch(op)
    {
	IMATH_FUNC_FOLDING(SIN, sin);
	IMATH_FUNC_FOLDING(COS, cos);
	IMATH_FUNC_FOLDING(TAN, tan);
	IMATH_FUNC_FOLDING(ASIN, asin);
	IMATH_FUNC_FOLDING(ACOS, acos);
	IMATH_FUNC_FOLDING(ATAN, atan);
	IMATH_FUNC_FOLDING(SINH, sinh);
	IMATH_FUNC_FOLDING(COSH, cosh);
	IMATH_FUNC_FOLDING(TANH, tanh);
	IMATH_FUNC_FOLDING(ASINH, asinh);
	IMATH_FUNC_FOLDING(ACOSH, acosh);
	IMATH_FUNC_FOLDING(ATANH, atanh);
	IMATH_FUNC_FOLDING(ABS, abs);
	IMATH_FUNC_FOLDING(FLOOR, floor);
	IMATH_FUNC_FOLDING(CEIL, ceil);
	IMATH_FUNC_FOLDING(EXP, exp);
	IMATH_FUNC_FOLDING(LOG, log);
	IMATH_FUNC_FOLDING(SQRT, sqrt);

      case ARG:
	type = T_CREAL;
	dval = 0.0;
	break;

      case LOG10:
	type = T_CREAL;
	dval = log((double)left) / M_LN10;
	break;

      case LOG2:
	type = T_CREAL;
	dval = log((double)left) / M_LN2;
	break;

      case INT_OP:
	type = T_CINT;
	ival = left;
	break;

      case REAL_OP:
	type = T_CREAL;
	dval = (double)left;
	break;

	ICONST_FOLDING(UMINUS, -);
	ICONST_FOLDING(NOT, !);
	ICONST_FOLDING(NOTNOT, !!);
      case ZERO:
	type = T_CINT;
	ival = 0;
	break;
      case ONE:
	type = T_CINT;
	ival = 1;
	break;
      case SIZEOF_OP:
	type = T_CINT;
	ival = sizeof(int);
	break;
      default:
	fprintf(stderr,	"Systematic Error: Undefined u-op: `%d', in integer_constant_folding1().\n",
		op);
	qhexit(1);
    }
#undef CONST_FOLDING
#undef IMATH_FUNC_FOLDING

    if(type == T_CINT)
	sprintf(temp_buff, "%d", ival);
    else
	sprintf(temp_buff, DOUBLE_EXPRESS, dval);

    t->type = type;
    t->val  = install_string(temp_buff);

    return t;
}

static Tree* real_constant_folding2(Tree* t)
{
    int op;
    int type = 0;
    double dval = 0;
    int ival = 0;
    double left, right;

    op = t->val;
    left  = atof(STRING(t->left->val));
    right = atof(STRING(t->right->val));
    reuse_tree(t->left), t->left = NULL;
    reuse_tree(t->right), t->right = NULL;

    switch(op)
    {
      case ADD:
	dval = left + right;
	type = T_CREAL;
	break;

      case SUB:
	dval = left - right;
	type = T_CREAL;
	break;

      case MUL:
	dval = left * right;
	type = T_CREAL;
	break;

      case DIV:
	dval = left / right;
	type = T_CREAL;
	break;

      case REM:
	dval = fmod(left, right);
	type = T_CREAL;
	break;

      case POW:
	dval = pow(left, right);
	type = T_CREAL;
	break;

      case OR:
	ival = left || right;
	type = T_CINT;
	break;

      case AND:
	ival = left && right;
	type = T_CINT;
	break;

      case EQ:
	ival = left == right;
	type = T_CINT;
	break;

      case NQ:
	ival = left != right;
	type = T_CINT;
	break;

      case LE:
	ival = left <= right;
	type = T_CINT;
	break;

      case LT:
	ival = left < right;
	type = T_CINT;
	break;

      case GE:
	ival = left >= right;
	type = T_CINT;
	break;

      case GT:
	ival = left > right;
	type = T_CINT;
	break;

      case LLSHIFT:
	ival = (unsigned int)left << (int)right;
	type = T_CINT;
	break;

      case LRSHIFT:
	ival = (unsigned int)left >> (int)right;
	type = T_CINT;
	break;

      case ALSHIFT:
	ival = (int)left << (int)right;
	type = T_CINT;
	break;

      case ARSHIFT:
	ival = (int)left >> (int)right;
	type = T_CINT;
	break;

      case ATAN2:
	dval = atan2(left, right);
	type = T_CREAL;
	break;

      default:
	fprintf(stderr, "Systematic Error: Undefined bin-op: `%d', in real_constant_folding2().\n", op);
	qhexit(1);
    }

    if(type == T_CINT)
	sprintf(temp_buff, "%d", ival);
    else
	sprintf(temp_buff, DOUBLE_EXPRESS, dval);

    t->type = type;
    t->val  = install_string(temp_buff);

    return t;
}

static Tree* real_constant_folding1(Tree* t)
{
    int op;
    int type = 0;
    double dval = 0;
    int ival = 0;
    double left;

    op = t->val;
    left  = atof(STRING(t->left->val));
    reuse_tree(t->left), t->left = NULL;

    switch(op)
    {
      case UMINUS:
	dval = -left;
	type = T_CREAL;
	break;
      case NOT:
	ival = !left;
	type = T_CINT;
	break;
      case NOTNOT:
	ival = !!left;
	type = T_CINT;
	break;
      case ZERO:
	ival = 0;
	type = T_CINT;
	break;
      case ONE:
	ival = 1;
	type = T_CINT;
	break;
      case SIZEOF_OP:
	ival = sizeof(double);
	type = T_CINT;
	break;

#define RMATH_FUNC_FOLDING(s, func) case s: type = T_CREAL; dval = func(left); break

	RMATH_FUNC_FOLDING(SIN,  sin   );
	RMATH_FUNC_FOLDING(COS,  cos   );
	RMATH_FUNC_FOLDING(TAN,  tan   );
	RMATH_FUNC_FOLDING(ASIN, asin  );
	RMATH_FUNC_FOLDING(ACOS, acos  );
	RMATH_FUNC_FOLDING(ATAN, atan  );
	RMATH_FUNC_FOLDING(SINH, sinh  );
	RMATH_FUNC_FOLDING(COSH, cosh  );
	RMATH_FUNC_FOLDING(TANH, tanh  );
	RMATH_FUNC_FOLDING(ASINH,asinh );
	RMATH_FUNC_FOLDING(ACOSH,acosh );
	RMATH_FUNC_FOLDING(ATANH,atanh );
	RMATH_FUNC_FOLDING(ABS,  abs   );
	RMATH_FUNC_FOLDING(FLOOR,floor );
	RMATH_FUNC_FOLDING(CEIL, ceil  );
	RMATH_FUNC_FOLDING(EXP,  exp   );
	RMATH_FUNC_FOLDING(LOG,  log   );
	RMATH_FUNC_FOLDING(SQRT, sqrt  );
#undef RMATH_FUNC_FOLDING

      case ARG:
	type = T_CREAL;
	dval = 0.0;
	break;

      case INT_OP:
	type = T_CINT;
	ival = (int)left;
	break;

      case REAL_OP:
	type = T_CREAL;
	dval = left;
	break;

      case LOG10:
	type = T_CREAL;
	dval = log(left) / M_LN10;
	break;

      case LOG2:
	type = T_CREAL;
	dval = log(left) / M_LN2;
	break;

      default:
	fprintf(stderr, "Systematic Error: Undefined bin-op: `%d', in real_constant_folding2().\n", op);
	qhexit(1);
    }

    if(type == T_CINT)
	sprintf(temp_buff, "%d", ival);
    else
	sprintf(temp_buff, DOUBLE_EXPRESS, dval);

    t->type = type;
    t->val  = install_string(temp_buff);

    return t;
}

#define IS_ZERO_NODE(t) (((t)->type == T_CINT && (t)->val == Q_zero_int) || \
			 ((t)->type == T_CREAL && atof(STRING((t)->val)) == 0.0))


/* const || const
   const || A
   A     || const,
*/
static Tree* or_constant_folding(Tree* t)
{
    int zero_left;
    int zero_right;
    int const_left;
    int const_right;

    zero_left  = IS_ZERO_NODE(t->left);
    zero_right = IS_ZERO_NODE(t->right);
    const_left = BETWEEN(T_CINT, t->left->type, T_CSTR);
    const_right = BETWEEN(T_CINT, t->right->type, T_CSTR);

    /* const || const */
    if(const_left && const_right)
    {
	reuse_tree(t->left);
	reuse_tree(t->right);
	t->left = NULL;
	t->right = NULL;
	t->type = T_CINT;
	if(zero_left && zero_right)
	    t->val = Q_zero_int;
	else
	    t->val = Q_one_int;
	return t;
    }

    /* const || A */
    if(const_left)
    {

	if(zero_left)		/* (0 || A) => notnot(A) */
	{
	    reuse_tree(t->left);
	    t->left = t->right;
	    t->right = NULL;
	    t->type = T_OP;
	    t->val = NOTNOT;
	    t->optimize_flag = False;
	}
	else			/* (nonzero || A) => 1 */
	{
	    reuse_tree(t->left);
	    reuse_tree(t->right);
	    t->left = NULL;
	    t->right = NULL;
	    t->type = T_CINT;
	    t->val = Q_one_int;
	}
	return t;
    }

    /* A || const */
    reuse_tree(t->right);
    t->right = NULL;
    t->type = T_OP;

    if(zero_right)		/* (A || 0) => notnot(A) */
	t->val = NOTNOT;
    else			/* (A || nonzero) => one(A) */
	t->val = ONE;
    return t;
}

/* const && const
   const && A
   A     && const,
*/
static Tree* and_constant_folding(Tree* t)
{
    int zero_left;
    int zero_right;
    int const_left;
    int const_right;

    zero_left  = IS_ZERO_NODE(t->left);
    zero_right = IS_ZERO_NODE(t->right);
    const_left = BETWEEN(T_CINT, t->left->type, T_CSTR);
    const_right = BETWEEN(T_CINT, t->right->type, T_CSTR);

    /* const && const */
    if(const_left && const_right)
    {
	reuse_tree(t->left);
	reuse_tree(t->right);
	t->left = NULL;
	t->right = NULL;
	t->type = T_CINT;
	if(zero_left || zero_right)
	    t->val = Q_zero_int;
	else
	    t->val = Q_one_int;
	return t;
    }

    /* const && A */
    if(const_left)
    {

	if(zero_left)		/* (0 && A) => 0 */
	{
	    reuse_tree(t->left);
	    reuse_tree(t->right);
	    t->left = NULL;
	    t->right = NULL;
	    t->type = T_CINT;
	    t->val = Q_zero_int;
	}
	else			/* (nonzero && A) => notnot(A) */
	{
	    reuse_tree(t->left);
	    t->left = t->right;
	    t->right = NULL;
	    t->type = T_OP;
	    t->val = NOTNOT;
	    t->optimize_flag = False;
	}
	return t;
    }

    /* A && const */
    reuse_tree(t->right);
    t->right = NULL;
    t->type = T_OP;

    if(zero_right)		/* (A && 0) => zero(A) */
	t->val = ZERO;
    else			/* (A && nonzero) => notnot(A) */
	t->val = NOTNOT;
    return t;
}

static Tree* optimize_tree1(Tree* t, Bool* is_changed)
{
    Tree* ltree = t->left;
    Tree* rtree = t->right;

    if(ltree != NULL && ltree->optimize_flag == False)
	ltree = t->left = optimize_tree1(ltree, is_changed);
    if(rtree != NULL && rtree->optimize_flag == False)
	rtree = t->right = optimize_tree1(rtree, is_changed);

    if(t->type == T_OP)
    {
	if(IS_2OP(t->val))
	{

	    if(t->val == COMPLEX_OP ||
	       t->val == RAND2)
	    {
		t->left  = optimize_tree1(t->left,  is_changed);
		t->right = optimize_tree1(t->right, is_changed);
		if(!*is_changed)
		{
		    t->optimize_flag = True;
		    t->left->optimize_flag = True;
		    t->right->optimize_flag = True;
		}
		return t;
	    }

	    if(ltree ->type == T_CINT &&
	       rtree->type == T_CINT)
	    {
		*is_changed = 1;
		return integer_constant_folding2(t);
	    }
	    if((ltree->type == T_CREAL && rtree->type == T_CREAL) ||
	       (ltree->type == T_CREAL && rtree->type == T_CINT) ||
	       (ltree->type == T_CINT  && rtree->type == T_CREAL))
	    {
		*is_changed = 1;
		return real_constant_folding2(t);
	    }
	    if(t->val == OR)
	    {
	        if(BETWEEN(T_CINT, ltree->type, T_CSTR) ||
		   BETWEEN(T_CINT, rtree->type, T_CSTR))
		{
		    *is_changed = 1;
		    return optimize_tree1(or_constant_folding(t), is_changed);
		}
		if(!*is_changed)
		    t->optimize_flag = True;
		return t;
	    }
	    if(t->val == AND)
	    {
		if(BETWEEN(T_CINT, ltree->type, T_CSTR) ||
		   BETWEEN(T_CINT, rtree->type, T_CSTR))
		{
		    *is_changed = 1;
		    return optimize_tree1(and_constant_folding(t), is_changed);
		}
		if(!*is_changed)
		    t->optimize_flag = True;
		return t;
	    }

	    if(t->val == ADD)
	    {
		/* 0 + t */
		if(ltree->type == T_CINT && ltree->val == Q_zero_int)
		{
		    Tree* tmp;

		    *is_changed = 1;
		    tmp = t;
		    t = t->right;
		    tmp->right = NULL;
		    reuse_tree(tmp);
		    t->optimize_flag = True;
		    return t;
		}

		/* t + 0*/
		if(rtree->type == T_CINT && rtree->val == Q_zero_int)
		{
		    Tree* tmp;

		    *is_changed = 1;
		    tmp = t;
		    t = t->left;
		    tmp->left = NULL;
		    reuse_tree(tmp);
		    t->optimize_flag = True;
		    return t;
		}
	    }
	    /* other's optimizations :
	       t - 0		=> t
	       t * 0		=> 0
	       t * 1		=> t
	       t / 1		=> t
	       t % 1		=> t
	       0 + t		=> t
	       0 - t		=> -t
	       0 * t		=> 0
	       1 * t		=> t
	       0 / t		=> 0
	       0 % t		=> 0
	       */
	}
	else if(IS_1OP(t->val))
	{
	    if(t->val == RAND0 ||
	       t->val == RAND1 ||
	       t->val == INDIR ||
	       t->val == IMAGE_OP ||
	       t->val == CONJUGATE ||
	       t->val == GETLINEOP)
	    {
		t->left  = optimize_tree1(t->left,  is_changed);
		if(!*is_changed)
		{
		    t->optimize_flag = True;
		    t->left->optimize_flag = True;
		}
		return t;
	    }


	    /*   op
	     *  /
	     * C
	     */
	    if(ltree->type == T_CINT)
	    {
		*is_changed = 1;
		return integer_constant_folding1(t);
	    }
	    if(ltree->type == T_CREAL)
	    {
		*is_changed = 1;
		return real_constant_folding1(t);
	    }
	    if(ltree->type == T_CSTR)
	    {
		switch(t->val)
		{
		  case INT_OP:
		    *is_changed = 1;
		    sprintf(temp_buff, "%d", atoi(STRING(ltree->val)));
		    t->type = T_CINT;
		    t->val = install_string(temp_buff);
		    reuse_tree(t->left);
		    t->left = NULL;
		    t->optimize_flag = True;
		    return t;
		  case REAL_OP:
		    *is_changed = 1;
		    sprintf(temp_buff, DOUBLE_EXPRESS, atof(STRING(ltree->val)));
		    t->type = T_CREAL;
		    t->val = install_string(temp_buff);
		    reuse_tree(t->left);
		    t->left = NULL;
		    t->optimize_flag = True;
		    return t;
		}
	    }

	    /* other's optimizations:
	       not:
	       not(not(t))	=> notnot(t)
	       not(notnot(t))	=> not(t)
	       not(one(t))	=> zero(t)
	       not(zero(t))	=> one(t)
	       not("A")		=> 0

	       notnot:
	       notnot("A")	=> 1
	       notnot(not(t))	=> not(t)
	       notnot(zero(t))	=> zero(t)
	       notnot(one(t))	=> one(t)

	       zero:
	       zero(U(t))	=> zero(t)
	       zero( nonfcall )	=> 0

	       one:
	       one(U(t))	=> one(t)
	       one( nonfcall )	=> 1

	       U(t) := unary operator
	       noncall : There is not FCALL in that tree
	     */
	  }
	/* other's optimizations :
	   t += 0	=> t
	   t -= 0	=> t
	   t *= 1	=> t
	   t /= 1	=> t
	   t %= 1	=> t
	   */
    }
    else if(t->type == T_VAR)
    {
	Symbol* enum_symbol;

	t->optimize_flag = True;

	enum_symbol = search_enumerate(t->val, t->module);
	if(enum_symbol != NULL)
	{
	    *is_changed = True;
	    t->type     = T_CINT;
	    t->val      = enum_symbol->name;
	    t->module   = None;
	    return t;
	}
    }

    if(!*is_changed)
	t->optimize_flag = True;
    return t;
}

Tree* optimize_tree(Tree* t)
{
    Bool is_changed;

    if(t == NULL)
	return NULL;

    if(!expr_optimazation)
	return t;

    do
    {
	is_changed = 0;
	t = optimize_tree1(t, &is_changed);
    }
    while(is_changed);

    return t;
}

static void non_constant_expr(void)
{
    yyerror("None constant expression");
}

Tree* calc_constant_expr_tree(Tree* expr)
{
    Bool is_changed;

    if(expr == NULL)
    {
	non_constant_expr();
	return NULL;
    }

    if(BETWEEN(T_CINT, expr->type, T_CSTR))
	return expr;

    do
    {
	is_changed = 0;
	expr = optimize_tree1(expr, &is_changed);
    }
    while(is_changed);

    if(!BETWEEN(T_CINT, expr->type, T_CSTR))
    {
	non_constant_expr();
	reuse_tree(expr);
	return NULL;
    }

    return expr;
}

Bool is_constant_variable_in_expr(long name, long module, Tree* t)
{
    if(t == NULL)
	return True;

    if(t->type == T_OP)
    {
	if(IS_ASSIGNOP(t->val) || IS_INCDECOP(t->val))
	{
	    if(t->left->type == T_VAR &&
	       t->left->val == name &&
	       t->left->module == module)
		return False;

	    if(t->left->type == T_ARRAY)
	    {
		if(t->left->left->type == T_VAR &&
		   t->left->left->val == name &&
		   t->left->left->module == module)
		    return False;
	    }
	}
    }

    return is_constant_variable_in_expr(name, module, t->left) &&
	   is_constant_variable_in_expr(name, module, t->right);
}
