/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "qhobj.h"
#include "optmz_tree.h"
#include "qhstruct.h"
#if defined(sgi)
#include <strings.h>
#endif

#define SORT_COMMAND "sort"

static Symbol* free_symbol_list = NULL;
extern int current_scope;
extern long current_module;

Symbol* global_symbol[SYMTABSIZ];
Symbol* local_symbol[SYMTABSIZ];
Symbol* local_label_symbol[SYMTABSIZ];
Symbol* global_label_symbol[SYMTABSIZ];
Symbol* constant_symbol[SYMTABSIZ];
Symbol* function_symbol[SYMTABSIZ];
Symbol* keyword_symbol[SYMTABSIZ];
Symbol* local_stmt_func_symbol[SYMTABSIZ];
Symbol* global_stmt_func_symbol[SYMTABSIZ];
Symbol* local_struct_symbol[SYMTABSIZ];
Symbol* global_struct_symbol[SYMTABSIZ];
Symbol* local_enum_symbol[SYMTABSIZ];
Symbol* global_enum_symbol[SYMTABSIZ];
Symbol* null_symbol;

static struct SymbolList* symbol_list_buffer = NULL;

static void clear_global_stmt_func_symbol(void);
static Symbol* search_stmt_func_tree(Symbol** stmt_func_symbol,
				     long name, int narg, long module);

static struct SymbolList* alloc_symbol_list(void)
{
    struct SymbolList* p;

    if(symbol_list_buffer != NULL)
    {
	p = symbol_list_buffer;
	symbol_list_buffer = symbol_list_buffer->next;
    }
    else
    {
	p = (struct SymbolList *)xmalloc(sizeof(struct SymbolList));
    }
    p->sym = NULL;
    p->next = NULL;

    return p;
}

static void reuse_symbol_list(struct SymbolList* p)
{
    p->next = symbol_list_buffer;
    symbol_list_buffer = p;
}

static void reuse_symbol_list_all(struct SymbolList* p)
{
    struct SymbolList* x;

    if(p == NULL)
	return;
    x = p;

    while(x->next)
	x = x->next;
    x->next = symbol_list_buffer;
    symbol_list_buffer = p;
}

static void delete_symbol_list(struct SymbolList* p)
{
    struct SymbolList* tmp;

    while(p)
    {
	tmp = p;
	p = p->next;
	tmp->next = NULL;
	xfree(tmp);
    }
}

static void push_binded_symbol(Symbol* sym, Symbol* x)
{
    struct SymbolList* p;

    p = alloc_symbol_list();
    p->sym = x;
    p->next = sym->local_binded;
    sym->local_binded = p;
}

static void pop_binded_symbol(Symbol* sym)
{
    struct SymbolList* p;

    p = sym->local_binded;
    sym->local_binded = sym->local_binded->next;

    reuse_symbol_list(p);
}

static inline Symbol* top_binded_symbol(Symbol* sym)
{
    return sym->local_binded->sym;
}

static Symbol* alloc_symbol(void)
{
    Symbol* p;

    if(free_symbol_list)
    {
	p = free_symbol_list;
	free_symbol_list = free_symbol_list->next;
    }
    else
    {
	p = (Symbol *)xmalloc(sizeof(Symbol));
    }

    bzero((char *)p, sizeof(Symbol));

    return p;
}

void init_symbol(void)
{
    int i;

    for(i = 0; i < SYMTABSIZ; i++)
    {
	global_symbol[i]           = NULL;
	local_symbol[i]            = NULL;
	local_label_symbol[i]      = NULL;
	global_label_symbol[i]     = NULL;
	constant_symbol[i]         = NULL;
	function_symbol[i]         = NULL;
	keyword_symbol[i]          = NULL;
	local_stmt_func_symbol[i]  = NULL;
	global_stmt_func_symbol[i] = NULL;
	local_struct_symbol[i]     = NULL;
	global_struct_symbol[i]    = NULL;
	local_enum_symbol[i]     = NULL;
	global_enum_symbol[i]    = NULL;
    }

    null_symbol = alloc_symbol();
}

void reuse_symbol(Symbol* sym)
{
    Symbol* p;

    if(sym == NULL || sym->type == None)
	return;
    sym->type = None;

    for(p = sym; p->next != NULL; p = p->next)
	if(p->local_binded)
	    reuse_symbol_list_all(p->local_binded);

    p->next = free_symbol_list;
    free_symbol_list = sym;
}

Symbol* install_variable(long name, long module)
{
    unsigned int addr;
    Symbol* sym;

    addr = ((unsigned int)(name + module)) % SYMTABSIZ;

    if(module >= 0)
    {
	for(sym = local_symbol[addr]; sym != NULL; sym = sym->next)
	    if(sym->name   == name &&
	       sym->module == module)
		return sym->local_binded ? top_binded_symbol(sym) : sym;
    }

    return install_global_symbol(name, module);
}

static Symbol* install_symbol_to_table(Symbol** symlist, long name, int module)
{
    Symbol* sym;
    Symbol* prev;
    unsigned int addr;

    if(module == -1)
	module = 0;

    addr = ((unsigned int)(name + module)) % SYMTABSIZ;

    sym = prev = symlist[addr];
    while(sym != NULL)
    {
	if(sym->name   == name &&
	   sym->module == module)
	{
	    return sym;
	}
	prev = sym;
	sym = sym->next;
    }

    /* install new symbol */
    if(prev == NULL)		/* symlist[addr] == NULL */
	symlist[addr] = sym = alloc_symbol();
    else			/* symlist[addr] != NULL */
	sym = prev->next = alloc_symbol();

    sym->type		= None;
    sym->name		= name;
    sym->module		= module;
    sym->local_binded	= NULL;
    sym->ref_cnt	= 0;

    return sym;
}

Symbol* install_label(long name, long module)
{
    Symbol* sym;

    if(current_scope == SYM_LOCAL)
	sym = install_symbol_to_table(local_label_symbol, name, module);
    else
	sym = install_symbol_to_table(global_label_symbol, name, module);

    if(sym->local_binded)
	sym = top_binded_symbol(sym);

    if(sym->type == SYM_LABEL)
    {
	sprintf(temp_buff, "Redefined label `%s'",
		module_name_string(module, name));
	yyerror(temp_buff);
	return sym;
    }
    sym->type = SYM_LABEL;

    return sym;
}

Symbol* search_label(long name, long module)
{
    Symbol* sym;

    if(current_scope == SYM_LOCAL)
	sym = install_symbol_to_table(local_label_symbol, name, module);
    else
	sym = install_symbol_to_table(global_label_symbol, name, module);

    if(sym->local_binded)
	sym = top_binded_symbol(sym);

    if(sym->type == None)
	sym->data.label_ref = NULL;

    return sym;
}

Symbol* install_local_symbol(long name, long module)
{
    Symbol* sym;

    sym = install_symbol_to_table(local_symbol, name, module);

    if(sym->local_binded)
	sym = top_binded_symbol(sym);

    sym->type       = SYM_VAR;
    sym->scope      = SYM_LOCAL;
    sym->data.align = 0;

    return sym;
}

Symbol* install_global_symbol(long name, long module)
{
    Symbol* sym;

    sym = install_symbol_to_table(global_symbol, name, module);
    if(sym->local_binded)
	sym = top_binded_symbol(sym);

    if(Wall_flag)
    {
	if(current_scope == SYM_LOCAL &&
	   sym->referenced_flag == False &&
	   sym->type == None &&
	   STRING(sym->name)[0] != temp_prefix)
	{
	    sprintf(temp_buff, "`%s' appear first global variable in the function",
		    module_name_string(sym->module, sym->name));
	    write_warning(temp_buff);
	}
    }

    if(sym->type == None)
    {
	sym->type       = SYM_VAR;
	sym->scope      = SYM_GLOBAL;
	sym->data.align = new_align();
	sym->referenced_flag = True;
    }

    return sym;
}

Symbol* newtemp_symbol(void)
{
    Symbol* sym;
    long name;

    name = newtemp();
    if(current_scope == SYM_GLOBAL)
	sym = install_global_symbol(name, current_module);
    else
	sym = install_local_symbol(name, current_module);

    if(sym->type == None)
    {
	sym->type = SYM_VAR;
	if(current_scope == SYM_GLOBAL)
	    sym->data.align = new_align();
	else
	    sym->data.align = 0;
    }

    return sym;
}

Symbol* install_constant_symbol(long name, int symtype)
{
    Symbol* sym;
    Symbol* prev;
    unsigned int addr;

    addr = name % SYMTABSIZ;

    sym = prev = constant_symbol[addr];
    while(sym != NULL)
    {
	if(sym->name == name &&
	   sym->type == symtype)
	    return sym;
	prev = sym;
	sym = sym->next;
    }

    /* install new symbol */
    if(prev == NULL)		/* constant_symbol[addr] == NULL */
	constant_symbol[addr] = sym = alloc_symbol();
    else			/* constant_symbol[addr] != NULL */
	sym = prev->next = alloc_symbol();

    sym->name		= name;
    sym->module		= None;
    sym->local_binded	= NULL;
    sym->type		= symtype;
    sym->ref_cnt	= 0;
    sym->data.align	= new_align();

    return sym;
}

Symbol* install_integer(int val)
{
    sprintf(temp_buff, "%d", val);
    return install_constant_symbol(install_string(temp_buff), SYM_INT);
}

char* symbol_name(Symbol* sym)
{
    static char buf[256];

    buf[0] = '\0';

    if(sym == NULL || sym == null_symbol)
	strcpy(buf, "NULL");
    else
    {
	switch(sym->type)
	{
	  case SYM_VAR:
	    sprintf(buf, "%s", module_name_string(sym->module, sym->name));
	    break;
	  case SYM_INT:
	  case SYM_REAL:
	    strcpy(buf, STRING(sym->name));
	    break;
	  case SYM_STR:
	    sprintf(buf, "\"%s\"", STRING(sym->name));
	    break;
	  case SYM_LABEL:
	    sprintf(buf, "%s:", module_name_string(sym->module, sym->name));
	    break;
	  case SYM_FUNC:
	    sprintf(buf, "%s/%d", module_name_string(sym->module, sym->name),
		    FUNCTAB(sym->data.func).narg);
	    break;
	  case SYM_KEYWORD:
	    switch(sym->name)
	    {
	      case QH_Key_Null:
		strcpy(buf, "key:NULL");
		break;
	      default:
		strcpy(buf, "<NULL>");
		break;
	    }
	    break;
	  default:
	    strcpy(buf, "<NULL>");
	    break;
	}
    }

    return buf;
}

Symbol* search_stmt_function(Tree* func)
{
    Symbol* sym;

    if(func->ftype == FUNC_GLOBAL)
	return NULL;

    if(func->ftype == FUNC_LOCAL_STFN)
	return search_stmt_func_tree(local_stmt_func_symbol,
				     func->left->val, func->val, func->left->module);

    if(func->ftype == FUNC_GLOBAL_STFN)
	return search_stmt_func_tree(global_stmt_func_symbol,
				     func->left->val, func->val, func->left->module);

    if((sym = search_stmt_func_tree(local_stmt_func_symbol,
				    func->left->val, func->val, func->left->module)) != NULL)
	return sym;
    if((sym = search_stmt_func_tree(global_stmt_func_symbol,
				    func->left->val, func->val, func->left->module)) != NULL)
	return sym;

    return NULL;
}

Symbol* install_function(long name, long module, int narg)
{
    Symbol* sym;
    Symbol* prev;
    unsigned int addr;

    if(module == -1)		/* function global::func */
	module = 0;

    addr = ((unsigned int)(name + module)) % SYMTABSIZ;

    sym = prev = function_symbol[addr];

    while(sym != NULL)
    {
	if(((sym->name                    == name   &&
	    sym->module                   == module &&
	    (FUNCTAB(sym->data.func).narg == narg)) ||
	    FUNCTAB(sym->data.func).extern_function == True))
	    return sym;

	prev = sym;
	sym = sym->next;
    }

    /* install new symbol */
    if(prev == NULL)		/* function_symbol[addr] == NULL */
	function_symbol[addr] = sym = alloc_symbol();
    else
	sym = prev->next = alloc_symbol();

    if(sym->type == None)
    {
	sym->name          = name;
	sym->module	   = module;
	sym->local_binded  = NULL;
	sym->type          = SYM_FUNC;
	sym->ref_cnt	   = 0;
	sym->data.func     = add_new_function(narg, name, module);
    }

    return sym;
}

void install_built_in_function(const char* name, int token, int narg)
{
    Symbol* sym;
    Function* f;

    sym = install_function(install_string(name), 0, narg);
    f = func_table + sym->data.func;

    f->addr      = -token;
    f->state     = F_Defined;
    f->offsp     = narg;
    f->load_flag = False;
    f->body      = null_quad_list();
}

static void delete_symbol(Symbol* sym)
{
    Symbol* p;

    while(sym)
    {
	p = sym;
	sym = sym->next;
	if(p->local_binded)
	    delete_symbol_list(p->local_binded);
	xfree(p);
    }
}

void delete_all_symbol(void)
{
    int i;

    delete_symbol_list(symbol_list_buffer);
    symbol_list_buffer = NULL;
    for(i = 0; i < SYMTABSIZ; i++)
    {
	delete_symbol(global_symbol[i]);
	delete_symbol(local_symbol[i]);
	delete_symbol(local_label_symbol[i]);
	delete_symbol(global_label_symbol[i]);
	delete_symbol(constant_symbol[i]);
	delete_symbol(function_symbol[i]);
	delete_symbol(keyword_symbol[i]);
	delete_symbol(local_enum_symbol[i]);
	delete_symbol(global_enum_symbol[i]);
    }

    clear_local_stmt_func_symbol();
    clear_global_stmt_func_symbol();
    clear_local_struct_symbol_table();
    clear_global_struct_symbol_table();
    delete_symbol(null_symbol);
    delete_symbol(free_symbol_list);
}

static void clear_symbol_table(Symbol** symlist)
{
    int i;

    for(i = 0; i < SYMTABSIZ; i++)
    {
	if(symlist[i])
	{
	    reuse_symbol(symlist[i]);
	    symlist[i] = NULL;
	}
    }
}

void clear_local_symbol_table(void)
{
    clear_symbol_table(local_symbol);
}

void clear_local_label_symbol_table(void)
{
    clear_symbol_table(local_label_symbol);
}

void clear_global_label_symbol_table(void)
{
    clear_symbol_table(global_label_symbol);
}

Symbol* install_keyword_symbol(long key)
{
    Symbol* sym;

    sym = install_symbol_to_table(keyword_symbol, key, None);
    if(sym->type == None)
    {
	sym->type = SYM_KEYWORD;
    }
    return sym;
}

void local_bind_symbol_name(long name, long module, Symbol* tmp)
{
    Symbol* sym;

    if(current_scope == SYM_LOCAL)
    {
	sym = install_symbol_to_table(local_symbol, name, module);
	if(sym->type == None)
	{
	    sym->type       = SYM_VAR;
	    sym->data.align = 0;
	}
    }
    else
    {
	sym = install_symbol_to_table(global_symbol, name, module);
	if(sym->type == None)
	{
	    sym->type       = SYM_VAR;
	    sym->data.align = new_align();
	}
    }

    /* Bind the symbol to new temporary variable */
    tmp->ref_cnt++;
    tmp->used_variable = False;
    push_binded_symbol(sym, tmp);
}

void local_unbind_symbol_name(long name, long module)
{
    Symbol* tmp;
    Symbol* sym;

    if(current_scope == SYM_LOCAL)
	sym = install_symbol_to_table(local_symbol, name, module);
    else
	sym = install_symbol_to_table(global_symbol, name, module);

    if(sym->local_binded == NULL)
    {
	fprintf(stderr, "Systematic error: Already unbinded symbol: "
		"`%s', in function local_unbind_symbol_name()\n",
		symbol_name(sym));
	qhexit(1);
    }

    if(W_unused_local_variable == True)
    {
	if(sym->local_binded->sym->used_variable == False &&
	   STRING(sym->name)[0] != temp_prefix)
	{
	    sprintf(temp_buff, "unused local variable `%s'", symbol_name(sym));
	    write_warning(temp_buff);
	}
	sym->local_binded->sym->used_variable = True;
	sym->used_variable = True;
    }

    /* remove binded variable */

    tmp = top_binded_symbol(sym);
    tmp->ref_cnt--;
    reuse_if_temp_sym(tmp);
    pop_binded_symbol(sym);
}

static void clear_stmt_func_symbol(Symbol** symlis)
{
    int i;
    Symbol* sym;

    for(i = 0; i < SYMTABSIZ; i++)
    {
	sym = symlis[i];
	while(sym)
	{
	    reuse_tree(sym->data.stmtfn.func);
	    reuse_tree(sym->data.stmtfn.expr);
	    sym = sym->next;
	}
	if(symlis[i])
	{
	    reuse_symbol(symlis[i]);
	    symlis[i] = NULL;
	}
    }
}

void clear_local_stmt_func_symbol(void)
{
    clear_stmt_func_symbol(local_stmt_func_symbol);
}

void clear_global_stmt_func_symbol(void)
{
    clear_stmt_func_symbol(global_stmt_func_symbol);
}

static void set_ftype(Tree* t, long name, int narg, long module)
{
    if(t->type != T_FCALL)
    {
	if(t->left != NULL)
	    set_ftype(t->left, name, narg, module);
	if(t->right != NULL)
	    set_ftype(t->right, name, narg, module);
	return;
    }

    if(t->left->val == name &&
       t->val == narg &&
       t->left->module == module)
	t->ftype = FUNC_GLOBAL;
    else
    {
	if(search_stmt_func_tree(local_stmt_func_symbol,
				 t->left->val,
				 t->val,
				 t->left->module) != NULL)
	{
	    t->ftype = FUNC_LOCAL_STFN;
	}
	else if(search_stmt_func_tree(global_stmt_func_symbol,
				      t->left->val,
				      t->val,
				      t->left->module) != NULL)
	    t->ftype = FUNC_GLOBAL_STFN;
	else
	{
	    t->ftype = FUNC_GLOBAL;
	}
    }

    if(t->right != NULL)
	set_ftype(t->right, name, narg, module);
}

static Bool check_func_argument(Tree* param)
{
    while(param != NULL)
    {
	if(param->left->type != T_VAR)
	    return False;
	param = param->right;
    }
    return True;
}

static void anarysis_constant_stfn_argument(Tree* param, Tree* expr)
{
    while(param)
    {
	if(is_constant_variable_in_expr(param->left->val,
					param->left->module,
					expr))
	{
	    param->left->optimize_flag = True;
	}
	else
	{
	    param->left->optimize_flag = False;
	}
	param = param->right;
    }
}

void install_stmt_func_tree(Tree* func, Tree* expr)
{
    Symbol* sym;
    Symbol* prev;
    unsigned int addr;
    long name;
    int narg;
    long module;
    int defun_scope;

    if(check_func_argument(func->right) == False)
    {
	yyerror("Bad statement function parameter");
	return;
    }

    name = func->left->val;
    narg = func->val;
    module = func->left->module;
    func = copy_tree(func);
    expr = copy_tree(expr);

    set_ftype(expr, name, narg, module);
    anarysis_constant_stfn_argument(func->right, expr);

    /* check redefine function */

    if(module >= 0)
    {
	addr = ((unsigned int)(name + module)) % SYMTABSIZ;
	if(current_scope == SYM_LOCAL)
	{
	    prev = sym = local_stmt_func_symbol[addr];
	    defun_scope = SYM_LOCAL;
	}
	else
	{
	    prev = sym = global_stmt_func_symbol[addr];
	    defun_scope = SYM_GLOBAL;
	}
    }
    else
    {
	module = 0;
	addr = ((unsigned int)name) % SYMTABSIZ;
	prev = sym = global_stmt_func_symbol[addr];
	defun_scope = SYM_GLOBAL;
    }

    while(sym)
    {
	if(sym->name                  == name &&
	   sym->module                == module &&
	   sym->data.stmtfn.func->val == narg)
	{
	    sprintf(temp_buff, "Redefined statement function `%s/%d'",
		    module_name_string(module, name), narg);
	    yyerror(temp_buff);
	    return;
	}
	prev = sym;
	sym = sym->next;
    }

    /* install new function */
    if(prev == NULL)		/* function_symbol[addr] == NULL */
	if(defun_scope == SYM_LOCAL)
	    local_stmt_func_symbol[addr] = sym = alloc_symbol();
	else
	    global_stmt_func_symbol[addr] = sym = alloc_symbol();
    else
	sym = prev->next = alloc_symbol();

    sym->name         = name;
    sym->module	      = module;
    sym->local_binded = NULL;
    sym->type         = SYM_STMT_FUNC;
    sym->ref_cnt      = 0;
    sym->data.stmtfn.func = func;
    sym->data.stmtfn.expr = expr;
}

static Symbol* search_stmt_func_tree(Symbol** symlis,
				     long name, int narg, long module)
{
    Symbol* sym;

    if(module == -1)		/* global::func */
	module = 0;

    sym = symlis[((unsigned int)(name + module)) % SYMTABSIZ];

    while(sym)
    {
	if(sym->name                  == name &&
	   sym->module                == module &&
	   sym->data.stmtfn.func->val == narg)
	{
	    return sym;		/* found */
	}
	sym = sym->next;
    }
    return NULL;		/* not found */
}

void print_all_global_symbol(void)
{
    int i;
    Symbol* sym;
    FILE* fp;

    if((fp = popen(SORT_COMMAND, "w")) == NULL)
    {
	fprintf(stderr, "Warning:");
	perror(SORT_COMMAND);
	fp = stdout;
    }

    for(i = 0; i < SYMTABSIZ; i++)
    {
	for(sym = global_symbol[i]; sym; sym = sym->next)
	{
	    if(STRING(sym->name)[0] != temp_prefix)
	    {
		if(sym->module)
		    fprintf(fp, "%s (%s)\n", STRING(sym->name), STRING(sym->module));
		else
		    fprintf(fp, "%s\n", STRING(sym->name));
	    }
	}
    }

    if(fp != stdout)
	pclose(fp);
}

void write_unused_local_variable(void)
{
    int i;
    Symbol* sym;

    if(!W_unused_local_variable)
	return;

    for(i = 0; i < SYMTABSIZ; i++)
    {
	sym = local_symbol[i];
	while(sym)
	{
	    if(sym->used_variable == False &&
	       STRING(sym->name)[0] != temp_prefix)
	    {
		sprintf(temp_buff, "unused local variable `%s'",
			module_name_string(sym->module, sym->name));
		write_warning(temp_buff);
	    }
	    sym = sym->next;
	}
    }
}

void set_initialized_flag(long name, long module)
{
    Symbol* sym;

    if(W_uninitialized_variable == False ||
       STRING(name)[0]          == temp_prefix)
	return;

    sym = install_variable(name, module);
    sym->initialized_flag = True;
}

void set_initialized_flag_all(Tree* t)
{
    if(t == NULL)
	return;

    if(t->type       == T_OP &&
       t->val        == ASSIGN &&
       t->left->type == T_VAR)
    {
	set_initialized_flag(t->left->val, t->left->module);
	set_initialized_flag_all(t->right);
	return;
    }

    set_initialized_flag_all(t->left);
    set_initialized_flag_all(t->right);
}

void declare_global_variable(long name, long module)
{
    Symbol* sym;

    sym = install_global_symbol(name, module);
    sym->declared_flag = True;
}

Symbol* search_struct_symbol(Tree* tag_node, Bool search_all)
{
    Symbol* sym;

    if(current_scope == SYM_LOCAL)
    {
	sym = install_symbol_to_table(local_struct_symbol,
				      tag_node->val,
				      tag_node->module);

	if(!search_all || sym->type != None)
	    return sym;
    }

    return install_symbol_to_table(global_struct_symbol,
				   tag_node->val,
				   tag_node->module);
}

static void clear_struct_symbol_table(Symbol** symlis)
{
    Symbol* sym;
    int i, entry;

    for(i = 0; i < SYMTABSIZ; i++)
    {
	sym = symlis[i];
	while(sym)
	{
	    if(sym->type != None)
	    {
		entry = sym->data.struct_entry;
		if(struct_tree[entry] != NULL)
		{
		    reuse_tree(struct_tree[entry]);
		    struct_tree[entry] = NULL;
		}
	    }
	    sym = sym->next;
	}
	if(symlis[i])
	{
	    reuse_symbol(symlis[i]);
	    symlis[i] = NULL;
	}
    }
}

void clear_local_struct_symbol_table(void)
{
    clear_struct_symbol_table(local_struct_symbol);
}

void clear_global_struct_symbol_table(void)
{
    clear_struct_symbol_table(global_struct_symbol);
}

void install_new_enumerator(long name, long module, int enumval)
{
    Symbol* sym;

    if(current_scope == SYM_GLOBAL)
	sym = install_symbol_to_table(global_enum_symbol, name, module);
    else
	sym = install_symbol_to_table(local_enum_symbol, name, module);

    if(sym->type != None)
    {
	sprintf(temp_buff, "redefinition enumerator of `%s'",
		module_name_string(module, name));
	yyerror(temp_buff);
	return;
    }

    sym->type = SYM_ENUM;
    sym->data.enum_val = install_integer(enumval);
}

void clear_local_enum_table(void)
{
    int i;

    for(i = 0; i < SYMTABSIZ; i++)
    {
	delete_symbol(local_enum_symbol[i]);
	local_enum_symbol[i] = NULL;
    }
}

Symbol* search_enumerate(long name, long module)
{
    Symbol* sym;

    if(module <= 0)
	sym = install_symbol_to_table(global_enum_symbol, name, module);
    else
	sym = install_symbol_to_table(local_enum_symbol, name, module);

    if(sym->type == None)
	return NULL;

    return sym->data.enum_val;
}
